Package: HiContacts
Title: Analysing cool files in R with HiContacts
Version: 1.13.0
Date: 2022-08-16
Authors@R: 
    c(person(given = "Jacques",
           family = "Serizay",
           role = c("aut", "cre"),
           email = "jacquesserizay@gmail.com",
           comment = c(ORCID = "0000-0002-4295-0624")))
Description: HiContacts provides a collection of tools to 
    analyse and visualize Hi-C datasets imported in R by HiCExperiment. 
License: MIT + file LICENSE
URL: https://github.com/js2264/HiContacts
BugReports: https://github.com/js2264/HiContacts/issues
Depends: R (>= 4.2), HiCExperiment
Imports: InteractionSet, SummarizedExperiment, GenomicRanges, IRanges,
        GenomeInfoDb, S4Vectors, methods, BiocGenerics, BiocIO,
        BiocParallel, RSpectra, Matrix, tibble, tidyr, dplyr, readr,
        stringr, ggplot2, ggrastr, scales, stats, utils
Suggests: HiContactsData, rtracklayer, GenomicFeatures, Biostrings,
        BSgenome.Scerevisiae.UCSC.sacCer3, WGCNA, Rfast, terra,
        patchwork, testthat (>= 3.0.0), BiocStyle, knitr, rmarkdown
biocViews: HiC, DNA3DStructure
Encoding: UTF-8
VignetteBuilder: knitr
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.3
git_url: https://git.bioconductor.org/packages/HiContacts
git_branch: devel
git_last_commit: 0b86ddb
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 22:20:06 UTC; biocbuild
Author: Jacques Serizay [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4295-0624>)
Maintainer: Jacques Serizay <jacquesserizay@gmail.com>

# Generated by roxygen2: do not edit by hand

export(Contacts)
export(afmhotrColors)
export(autocorrelate)
export(bbrColors)
export(bgrColors)
export(boost)
export(bwrColors)
export(cisTransRatio)
export(coarsen)
export(contacts_yeast)
export(contacts_yeast_eco1)
export(coolerColors)
export(despeckle)
export(detrend)
export(divide)
export(getBorders)
export(getCompartments)
export(getDiamondInsulation)
export(getLoops)
export(localDistanceLaw)
export(plot4C)
export(plotPs)
export(plotPsSlope)
export(plotSaddle)
export(plotScalogram)
export(rainbowColors)
export(scalogram)
export(subsample)
export(virtual4C)
exportMethods(aggregate)
exportMethods(coverage)
exportMethods(distanceLaw)
exportMethods(merge)
exportMethods(montage)
exportMethods(normalize)
exportMethods(plotMatrix)
import(HiCExperiment)
import(InteractionSet)
import(stringr)
import(tibble)
import(tidyr)
importFrom(BiocGenerics,normalize)
importFrom(BiocGenerics,path)
importFrom(BiocIO,import)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,countOverlaps)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,findOverlaps)
importFrom(GenomicRanges,resize)
importFrom(GenomicRanges,seqnames)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,tileGenome)
importFrom(IRanges,IRanges)
importFrom(IRanges,coverage)
importFrom(Matrix,sparseMatrix)
importFrom(RSpectra,eigs_sym)
importFrom(S4Vectors,SimpleList)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,queryHits)
importFrom(SummarizedExperiment,assay)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,cur_data)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_modify)
importFrom(dplyr,group_split)
importFrom(dplyr,lead)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,pull)
importFrom(dplyr,relocate)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,slice_max)
importFrom(dplyr,summarize)
importFrom(dplyr,tally)
importFrom(ggrastr,geom_tile_rast)
importFrom(readr,DataFrameCallback)
importFrom(readr,read_tsv_chunked)
importFrom(scales,oob_censor)
importFrom(scales,oob_squish)
importFrom(scales,rescale)
importFrom(scales,trans_breaks)
importFrom(scales,trans_format)
importFrom(scales,unit_format)
importFrom(stats,cor)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,drop_na)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,pivot_wider)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)

<!-- badges: start -->
[![DOI](https://img.shields.io/badge/Nat._Commun-10.1038/s41467--024--44761--x-blue)](https://www.nature.com/articles/s41467-024-44761-x)
[![](https://img.shields.io/badge/lifecycle-stable-green.svg)](https://www.tidyverse.org/lifecycle/#stable)
[![](https://img.shields.io/badge/license-MIT-green.svg)](https://opensource.org/licenses/MIT)
[![rworkflows](https://github.com/js2264/HiContacts/actions/workflows/rworkflows.yml/badge.svg)](https://github.com/js2264/HiContacts/actions/workflows/rworkflows.yml)
[![Documentation](https://github.com/js2264/HiContacts/workflows/pkgdown/badge.svg)](https://js2264.github.io/HiContacts)
[![OHCA book](https://github.com/js2264/OHCA/actions/workflows/pages/pages-build-deployment/badge.svg)](https://js2264.github.io/OHCA/)
<a href=http://bioconductor.org/packages/release/bioc/html/HiContacts.html><img alt="Static Badge" src="https://img.shields.io/badge/Bioc_(release)-Landing_page-green?link=http%3A%2F%2Fbioconductor.org%2FcheckResults%2Fdevel%2Fbioc-LATEST%2FHiContacts%2F"></a>
<a href=http://bioconductor.org/checkResults/release/bioc-LATEST/HiContacts/><img alt="Bioc build (release)" src="https://img.shields.io/badge/dynamic/yaml?url=https%3A%2F%2Fbioconductor.org%2FcheckResults%2Frelease%2Fbioc-LATEST%2FHiContacts%2Fraw-results%2Fnebbiolo1%2Fbuildsrc-summary.dcf&query=%24.Status&label=Bioc%20build%20(release)&link=https%3A%2F%2Fbioconductor.org%2FcheckResults%2Frelease%2Fbioc-LATEST%2FHiContacts%2F"></a>
<a href=http://bioconductor.org/checkResults/devel/bioc-LATEST/HiContacts/><img alt="Bioc build (devel)" src="https://img.shields.io/badge/dynamic/yaml?url=https%3A%2F%2Fbioconductor.org%2FcheckResults%2Fdevel%2Fbioc-LATEST%2FHiContacts%2Fraw-results%2Fnebbiolo2%2Fbuildsrc-summary.dcf&query=%24.Status&label=Bioc%20build%20(devel)&link=https%3A%2F%2Fbioconductor.org%2FcheckResults%2Fdevel%2Fbioc-LATEST%2FHiContacts%2F"></a>
<!-- badges: end -->

# HiContacts

[👉 OHCA book 📖](https://js2264.github.io/OHCA/)

*Please cite:* 

Serizay J, Matthey-Doret C, Bignaud A, Baudry L, Koszul R (2024). “Orchestrating chromosome conformation capture analysis with Bioconductor.” _Nature Communications_, **15**, 1-9. [doi:10.1038/s41467-024-44761-x](https://doi.org/10.1038/s41467-024-44761-x).

[![DOI](https://img.shields.io/badge/Nat._Commun-10.1038/s41467--024--44761--x-blue)](https://www.nature.com/articles/s41467-024-44761-x)

---

HiContacts provides tools to investigate `(m)cool` matrices imported in R by `HiCExperiment`. 

It leverages the `HiCExperiment` class of objects, built on pre-existing Bioconductor objects, namely `InteractionSet`, `GInterations` and `ContactMatrix` (`Lun, Perry & Ing-Simmons, F1000Research 2016`), and provides **analytical** and **visualization** tools to investigate contact maps. 

## Installation

`HiContacts` is available in Bioconductor. To install the current release, use:

```r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("HiContacts")
```

To install the most recent version of `HiContacts`, you can use:

```r
install.packages("devtools")
devtools::install_github("js2264/HiContacts")
library(HiContacts)
```

## Citation

If you are using `HiContacts` in your research, please cite: 

> Serizay J (2022). _HiContacts: HiContacts: R interface to cool files_.
> R package version 1.1.0
> <https://github.com/js2264/HiContacts>.

## How to use `HiContacts`

`HiContacts` includes a introduction vignette where its usage is 
illustrated. To access the vignette, please use:

```r
vignette('HiContacts')
```

## Visualising Hi-C contact maps and features 

### Importing a Hi-C contact maps file with `HiCExperiment`

```r
mcool_file <- HiContactsData::HiContactsData('yeast_wt', format = 'mcool')
range <- 'I:20000-80000' # range of interest
availableResolutions(mcool_file)
hic <- HiCExperiment::import(mcool_file, format = 'mcool', focus = range, resolution = 1000)
hic
```

### Plotting matrices (square or horizontal)

```r
plotMatrix(hic, use.scores = 'count')
plotMatrix(hic, use.scores = 'balanced', limits = c(-4, -1))
plotMatrix(hic, use.scores = 'balanced', limits = c(-4, -1), maxDistance = 100000)
```

### Plotting matrices with topological features

```r
library(rtracklayer)
mcool_file <- HiContactsData::HiContactsData('yeast_wt', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', focus = 'IV')
loops <- system.file("extdata", 'S288C-loops.bedpe', package = 'HiContacts') |> 
    import() |> 
    InteractionSet::makeGInteractionsFromGRangesPairs()
borders <- system.file("extdata", 'S288C-borders.bed', package = 'HiContacts') |> 
    import()
p <- plotMatrix(
    hic, loops = loops, borders = borders, 
    limits = c(-4, -1), dpi = 120
)
```

### Plotting aggregated matrices (a.k.a. APA plots) 

```r
contacts <- contacts_yeast()
contacts <- zoom(contacts, resolution = 2000)
aggr_centros <- aggregate(contacts, targets = topologicalFeatures(contacts, 'centromeres'))
plotMatrix(aggr_centros, use.scores = 'detrended', limits = c(-1, 1), scale = 'linear')
```

## Mapping topological features 

### Chromosome compartments 

```r
microC_mcool <- fourDNData::fourDNData('4DNES14CNC1I', 'mcool')
hic <- import(microC_mcool, format = 'mcool', resolution = 10000000)
genome <- BSgenome.Mmusculus.UCSC.mm10::BSgenome.Mmusculus.UCSC.mm10

# - Get compartments
hic <- getCompartments(
    hic, resolution = 100000, genome = genome, chromosomes = c('chr17', 'chr19')
)

# - Export compartments as bigwig and bed files
export(IRanges::coverage(metadata(hic)$eigens, weight = 'eigen'), 'microC_compartments.bw')
export(
    topologicalFeatures(hic, 'compartments')[topologicalFeatures(hic, 'compartments')$compartment == 'A'], 
    'microC_A-compartments.bed'
)
export(
    topologicalFeatures(hic, 'compartments')[topologicalFeatures(hic, 'compartments')$compartment == 'B'], 
    'microC_B-compartments.bed'
)

# - Generate saddle plot
plotSaddle(hic)
```

### Diamond insulation score and chromatin domains borders

```r
# - Compute insulation score
hic <- refocus(hic, 'chr19:1-30000000') |> 
    zoom(resolution = 10000) |> 
    getDiamondInsulation(window_size = 100000) |> 
    getBorders()

# - Export insulation as bigwig track and borders as bed file
export(IRanges::coverage(metadata(hic)$insulation, weight = 'insulation'), 'microC_insulation.bw')
export(topologicalFeatures(hic, 'borders'), 'microC_borders.bed')
```

## In-depth analysis of `HiCExperiment` objects

### Arithmetics

#### Detrend
#### Autocorrelate
#### Divide
#### Merge

### Distance law, a.k.a. P(s)

```r
hic <- import(CoolFile(
    mcool_file, 
    pairs = HiContactsData::HiContactsData('yeast_wt', format = 'pairs.gz')
))
ps <- distanceLaw(hic)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p))
```

### Virtual 4C

```r
hic <- import(CoolFile(mcool_file))
v4C <- virtual4C(hic, viewpoint = GRanges('V:150000-170000'))
plot4C(v4C)
```

### Cis-trans ratios

```r
hic <- import(CoolFile(mcool_file))
cisTransRatio(hic)
```

### Scalograms 

```r

```

## HiCExperiment ecosystem

`HiCool` is integrated within the `HiCExperiment` ecosystem in Bioconductor. 
Read more about the `HiCExperiment` class and handling Hi-C data in R 
[here](https://github.com/js2264/HiCExperiment).

![](https://raw.githubusercontent.com/js2264/HiCExperiment/devel/man/figures/HiCExperiment_ecosystem.png)

- [HiCExperiment](https://github.com/js2264/HiCExperiment): Parsing Hi-C files in R
- [HiCool](https://github.com/js2264/HiCool): End-to-end integrated workflow to process fastq files into .cool and .pairs files
- [HiContacts](https://github.com/js2264/HiContacts): Investigating Hi-C results in R
- [HiContactsData](https://github.com/js2264/HiContactsData): Data companion package
- [fourDNData](https://github.com/js2264/fourDNData): Gateway package to 4DN-hosted Hi-C experiments

## ----eval = TRUE, echo=FALSE, results="hide", warning=FALSE-------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>", 
    crop = NULL
)
suppressPackageStartupMessages({
    library(ggplot2)
    library(dplyr)
    library(GenomicRanges)
    library(HiCExperiment)
    library(HiContactsData)
    library(HiContacts)
})

## -----------------------------------------------------------------------------
citation('HiContacts')

## -----------------------------------------------------------------------------
library(dplyr)
library(ggplot2)
library(HiCExperiment)
library(HiContacts)
library(HiContactsData)
library(rtracklayer)
library(InteractionSet)
cool_file <- HiContactsData('yeast_wt', format = 'cool')
hic <- import(cool_file, format = 'cool')
hic

## -----------------------------------------------------------------------------
## Square matrix
plotMatrix(hic, use.scores = 'balanced', limits = c(-4, -1))

## Horizontal matrix
plotMatrix(
    refocus(hic, 'II'),
    use.scores = 'balanced', limits = c(-4, -1), 
    maxDistance = 200000
)

## -----------------------------------------------------------------------------
mcool_file <- HiContactsData('yeast_wt', format = 'mcool')
loops <- system.file("extdata", 'S288C-loops.bedpe', package = 'HiCExperiment') |> 
    import() |> 
    makeGInteractionsFromGRangesPairs()
p <- import(mcool_file, format = 'mcool', focus = 'IV') |> 
    plotMatrix(loops = loops, limits = c(-4, -1), dpi = 120)

## -----------------------------------------------------------------------------
borders <- system.file("extdata", 'S288C-borders.bed', package = 'HiCExperiment') |> 
    import()
p <- import(mcool_file, format = 'mcool', focus = 'IV') |> 
    plotMatrix(loops = loops, borders = borders, limits = c(-4, -1), dpi = 120)

## -----------------------------------------------------------------------------
aggr_centros <- HiContacts::aggregate(
    hic, targets = loops, BPPARAM = BiocParallel::SerialParam()
)
plotMatrix(
    aggr_centros, use.scores = 'detrended', limits = c(-1, 1), scale = 'linear', 
    cmap = bgrColors()
)

## -----------------------------------------------------------------------------
mcool_file <- HiContactsData('mESCs', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', focus = 'chr2', resolution = 160000)
hic <- autocorrelate(hic)
scores(hic)
summary(scores(hic, 'autocorrelated'))
plotMatrix(hic, use.scores = 'autocorrelated', limits = c(-1, 1), scale = 'linear')

## -----------------------------------------------------------------------------
hic <- import(mcool_file, format = 'mcool', focus = 'chr18:20000000-35000000', resolution = 40000)
detrended_hic <- detrend(hic)
patchwork::wrap_plots(
    plotMatrix(detrended_hic, use.scores = 'expected', scale = 'log10', limits = c(-3, -1), dpi = 120),
    plotMatrix(detrended_hic, use.scores = 'detrended', scale = 'linear', limits = c(-1, 1), dpi = 120)
)

## -----------------------------------------------------------------------------
mcool_file_1 <- HiContactsData('yeast_eco1', format = 'mcool')
mcool_file_2 <- HiContactsData('yeast_wt', format = 'mcool')
hic_1 <- import(mcool_file_1, format = 'mcool', focus = 'II:1-300000', resolution = 2000)
hic_2 <- import(mcool_file_2, format = 'mcool', focus = 'II:1-300000', resolution = 2000)
merged_hic <- merge(hic_1, hic_2) 
hic_1
hic_2
merged_hic

## -----------------------------------------------------------------------------
hic_1 <- import(mcool_file_1, format = 'mcool', focus = 'II', resolution = 2000)
hic_2 <- import(mcool_file_2, format = 'mcool', focus = 'II', resolution = 2000)
div_hic <- divide(hic_1, by = hic_2) 
div_hic
p <- patchwork::wrap_plots(
    plotMatrix(hic_1, use.scores = 'balanced', scale = 'log10', limits = c(-4, -1)),
    plotMatrix(hic_2, use.scores = 'balanced', scale = 'log10', limits = c(-4, -1)),
    plotMatrix(div_hic, use.scores = 'balanced.fc', scale = 'log2', limits = c(-2, 2), cmap = bwrColors())
)

## -----------------------------------------------------------------------------
hic_1_despeckled <- despeckle(hic_1)
hic_1_despeckled5 <- despeckle(hic_1, focal.size = 5)
p <- patchwork::wrap_plots(
    plotMatrix(hic_1, use.scores = 'balanced', scale = 'log10', limits = c(-4, -1)),
    plotMatrix(hic_1_despeckled, use.scores = 'balanced.despeckled', scale = 'log10', limits = c(-4, -1)),
    plotMatrix(hic_1_despeckled5, use.scores = 'balanced.despeckled', scale = 'log10', limits = c(-4, -1))
)

## -----------------------------------------------------------------------------
mcool_file <- HiContactsData('yeast_wt', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', resolution = 16000)

# - Get compartments
hic <- getCompartments(hic, chromosomes = c('XV', 'XVI'))
hic

# - Export compartments as bigwig and bed files
export(IRanges::coverage(metadata(hic)$eigens, weight = 'eigen'), 'compartments.bw')
export(
    topologicalFeatures(hic, 'compartments')[topologicalFeatures(hic, 'compartments')$compartment == 'A'], 
    'A-compartments.bed'
)
export(
    topologicalFeatures(hic, 'compartments')[topologicalFeatures(hic, 'compartments')$compartment == 'B'], 
    'B-compartments.bed'
)

# - Generate saddle plot
plotSaddle(hic)

## -----------------------------------------------------------------------------
# - Compute insulation score
hic <- refocus(hic, 'II:1-300000') |> 
    zoom(resolution = 1000) |> 
    getDiamondInsulation(window_size = 8000) |> 
    getBorders()
hic

# - Export insulation as bigwig track and borders as bed file
export(IRanges::coverage(metadata(hic)$insulation, weight = 'insulation'), 'insulation.bw')
export(topologicalFeatures(hic, 'borders'), 'borders.bed')

## -----------------------------------------------------------------------------
mcool_file <- HiContactsData('mESCs', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', focus = 'chr18:20000000-35000000', resolution = 40000)
v4C <- virtual4C(hic, viewpoint = GRanges('chr18:31000000-31050000'))
plot4C(v4C, ggplot2::aes(x = center, y = score))

## -----------------------------------------------------------------------------
mcool_file <- HiContactsData('yeast_wt', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', resolution = 1000)
cisTransRatio(hic)

## -----------------------------------------------------------------------------
# Without a pairs file
mcool_file <- HiContactsData('yeast_wt', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', resolution = 1000)
ps <- distanceLaw(hic)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p))

# With a pairs file
pairsFile(hic) <- HiContactsData('yeast_wt', format = 'pairs.gz')
ps <- distanceLaw(hic)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p))
plotPsSlope(ps, ggplot2::aes(x = binned_distance, y = slope))

# Comparing P(s) curves
c1 <- import(
    HiContactsData('yeast_wt', format = 'mcool'), 
    format = 'mcool',
    resolution = 1000, 
    pairsFile = HiContactsData('yeast_wt', format = 'pairs.gz')
)
c2 <- import(
    HiContactsData('yeast_eco1', format = 'mcool'), 
    format = 'mcool',
    resolution = 1000, 
    pairsFile = HiContactsData('yeast_eco1', format = 'pairs.gz')
)
ps_1 <- distanceLaw(c1) |> mutate(sample = 'WT')
ps_2 <- distanceLaw(c2) |> mutate(sample = 'Eco1-AID')
ps <- rbind(ps_1, ps_2)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p, group = sample, color = sample))
plotPsSlope(ps, ggplot2::aes(x = binned_distance, y = slope, group = sample, color = sample))

## -----------------------------------------------------------------------------
sessionInfo()

################################################################################
#                                                                              #
#                               CONSTRUCTORS                                   #
#                                                                              #
################################################################################

#' @title Contacts
#' 
#' @rdname Contacts
#' 
#' @description 
#' 
#' This function has been deprecated in favor of the generic `HiCExperiment()`
#' constructor (from HiCExperiment package).
#' 
#' @param file Path to a (m)cool file
#' @param resolution Resolution to use with mcool file
#' @param focus focus Chr. coordinates for which 
#'   interaction counts are extracted from the .(m)cool file, provided
#'   as a character string (e.g. "II:4001-5000"). If not provided, 
#'   the entire (m)cool file will be imported. 
#' @param metadata list of metadata
#' @param topologicalFeatures topologicalFeatures provided as a named SimpleList
#' @param pairsFile Path to an associated .pairs file
#' @return a new `HiCExperiment` object.
#' 
#' @import HiCExperiment
#' @export
#' @examples 
#' library(HiContacts)
#' library(HiContactsData)
#' mcool_path <- HiContactsData::HiContactsData('yeast_wt', 'mcool')
#' Contacts(mcool_path, resolution = 1000)

Contacts <- function(
    file, 
    resolution = NULL, 
    focus = NULL, 
    metadata = list(), 
    topologicalFeatures = S4Vectors::SimpleList(
        'loops' = S4Vectors::Pairs(
            GenomicRanges::GRanges(), 
            GenomicRanges::GRanges()
        ), 
        'borders' = GenomicRanges::GRanges(), 
        'compartments' = GenomicRanges::GRanges(), 
        'viewpoints' = GenomicRanges::GRanges()
    ), 
    pairsFile = NULL
) {
    .Deprecated(
            "HiCExperiment::HiCExperiment", 
            msg = paste(
                "`HiContacts::Contacts` is deprecated;", 
                "see '?HiCExperiment::HiCExperiment' constructor instead."
            )
    )
    HiCExperiment::HiCExperiment(
        file = file, 
        resolution = resolution, 
        focus = focus, 
        metadata = metadata, 
        topologicalFeatures = topologicalFeatures, 
        pairsFile = pairsFile
    )
} 

#' @importFrom stats rbinom
#' @rdname arithmetics
#' @export

subsample <- function(x, prop) {

    gis <- HiCExperiment::interactions(x)
    gis$score <- HiCExperiment::scores(x, 'count')
    gis$scaling <- gis$balanced/gis$score
    cm <- HiCExperiment::gi2cm(gis)
    an <- InteractionSet::anchors(cm)
    mat <- HiCExperiment::cm2matrix(cm)
    scores_names <- names(scores(x))
    if (prop > 1) prop <- prop/sum(mat, na.rm = TRUE)

    subsample_counts <- apply(mat, 2, function(x) stats::rbinom(nrow(mat), x, prop))
    subsample_counts[subsample_counts == 0] <- NA
    subsample_counts[lower.tri(subsample_counts)] <- NA
    # subsample_counts <- 2 * subsample_counts
    cm_subsampled <- InteractionSet::ContactMatrix(
        subsample_counts, 
        anchor1 = an[[1]], 
        anchor2 = an[[2]], 
        regions = InteractionSet::regions(cm)
    )
    is_subsampled <- InteractionSet::deflate(cm_subsampled)
    gis_subsampled <- InteractionSet::interactions(is_subsampled)
    gis_subsampled$bin_id1 <- HiCExperiment::anchors(gis_subsampled)[[1]]$bin_id
    gis_subsampled$bin_id2 <- HiCExperiment::anchors(gis_subsampled)[[2]]$bin_id
    HiCExperiment::interactions(x) <- gis_subsampled
    m <- dplyr::left_join(
        as.data.frame(GenomicRanges::mcols(gis_subsampled)), 
        as.data.frame(GenomicRanges::mcols(gis)), 
        by = c('bin_id1', 'bin_id2')
    ) |> dplyr::select(-bin_id1, -bin_id2, -score)
    n <- data.frame(
        count = SummarizedExperiment::assay(is_subsampled, 1)[, 1]
    )
    n$balanced <- n$count * m$scaling
    n$balanced <- scales::rescale(n$balanced, to = c(
        min(gis$balanced, na.rm = TRUE),
        max(gis$balanced, na.rm = TRUE)
    ))
    l <- as.list(n) |> S4Vectors::SimpleList()
    x@scores <- l

    return(x)

}

#' Compute a scalogram of contacts
#' 
#' @name scalogram
#' 
#' @param x A `HiCExperiment` object
#' @param dist_min Minimum distance for interactions to be considered.
#' @param nbins Number of bins to divide each chromosome
#' @param probs Quantiles of interactions 
#' @return a tibble
#' 
#' @return a tibble
#'
#' @importFrom tibble tibble
#' @importFrom tidyr drop_na
#' @importFrom tidyr pivot_longer
#' @importFrom dplyr select
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr group_by
#' @importFrom dplyr group_modify
#' 
#' @examples 
#' contacts_yeast <- HiCExperiment::contacts_yeast()
#' pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
#'   'yeast_wt', format = 'pairs.gz'
#' )
#' scalo <- scalogram(contacts_yeast['II'])
#' scalo
NULL

#' @rdname scalogram
#' @export

scalogram <- function(x, dist_min = 0, nbins = 250, probs = c(0.25, 0.5, 0.75)) {
    pairsFile <- HiCExperiment::pairsFile(x)
    if (is.null(pairsFile)) {
        stop("No PairsFile is associated with the provided HiCExperiment object.")
    }
    message("Importing pairs file ", pairsFile, " in memory. This may take a while...")
    pairs <- BiocIO::import(pairsFile, format = 'pairs')
    cispairs <- pairs[InteractionSet::intrachr(pairs)]
    an_ <- HiCExperiment::anchors(cispairs)
    scalo <- tibble::tibble(
        dist = cispairs$distance, 
        chr = as.vector(GenomicRanges::seqnames(an_[[1]])), 
        pos1 = GenomicRanges::start(an_[[1]]),
        pos2 = GenomicRanges::start(an_[[2]])
    ) |>
        tidyr::pivot_longer(-c(chr, dist), names_to = 'pos_', values_to = 'pos') |>
        dplyr::select(-pos_) |>
        dplyr::filter(dist > dist_min, pos > 1) 
    seq_ <- seq(1, max(scalo$pos, na.rm = TRUE), length.out = nbins + 1)
    df <- scalo |> 
        dplyr::mutate(binned_pos = seq_[findInterval(pos, vec = seq_)]) |> 
        dplyr::group_by(chr, binned_pos) |>
        dplyr::group_modify(~ {
            tibble::tibble(
                prob = factor(probs, rev(probs)), 
                dist_quantile = quantile(.x$dist, probs = probs)
            )
        }) 
    return(df)
}

#' @rdname arithmetics
#' @export

coarsen <- function(x, bin.size) {

    res <- resolution(x)
    if (bin.size <= res) stop("Provided `bin.size` <= resolution(x). Please coarsen only to a coarser resolution.")
    gis <- HiCExperiment::interactions(x)
    an <- anchors(gis)
    re <- regions(gis)
    rebinned_re <- GenomicRanges::tileGenome(
        GenomeInfoDb::seqinfo(re), tilewidth = bin.size, 
        cut.last.tile.in.chrom = TRUE
    )
    rebinned_re$chr <- as.vector(GenomicRanges::seqnames(rebinned_re))
    rebinned_re$center <- floor(as.vector(GenomicRanges::start(rebinned_re) + {GenomicRanges::end(rebinned_re) - GenomicRanges::start(rebinned_re)}/2))
    rebinned_re$bin_id <- seq_along(rebinned_re) - 1
    names(rebinned_re) <- paste(
        as.vector(GenomicRanges::seqnames(rebinned_re)), 
        as.vector(GenomicRanges::start(rebinned_re)), 
        as.vector(GenomicRanges::end(rebinned_re)), 
        sep = '_'
    )
    regions(gis)$bin_id <- rebinned_re[GenomicRanges::nearest(
        GenomicRanges::resize(regions(gis), fix = 'center', width = 1), rebinned_re
    )]$bin_id

    # - Coarsen GIs
    score_names <- names(scores(x))
    gis$bin_id1 <- rebinned_re[GenomicRanges::nearest(
        GenomicRanges::resize(an[[1]], fix = 'center', width = 1), 
        rebinned_re
    )]$bin_id
    gis$bin_id2 <- rebinned_re[GenomicRanges::nearest(
        GenomicRanges::resize(an[[2]], fix = 'center', width = 1), 
        rebinned_re
    )]$bin_id
    df <- as.data.frame(S4Vectors::mcols(gis))
    gis2 <- dplyr::select(df, bin_id1, bin_id2) |> 
        dplyr::distinct() |>
        dplyr::left_join(as.data.frame(rebinned_re), by = c('bin_id1' = 'bin_id')) |> 
        dplyr::select(seqnames, start, end, bin_id1, bin_id2) |> 
        dplyr::rename(seqnames1 = seqnames, start1 = start, end1 = end) |> 
        dplyr::left_join(as.data.frame(rebinned_re), by = c('bin_id2' = 'bin_id')) |> 
        dplyr::select(seqnames1, start1, end1, seqnames, start, end, bin_id1, bin_id2) |> 
        dplyr::rename(seqnames2 = seqnames, start2 = start, end2 = end) |> 
        HiCExperiment::df2gi()
    replaceRegions(gis2) <- rebinned_re

    # - Coarsen scores
    scores <- dplyr::group_by(df, bin_id1, bin_id2) |> 
        dplyr::summarise(dplyr::across(all_of(score_names), function(x)sum(x)), .groups = "drop")
    scores <- left_join(as.data.frame(S4Vectors::mcols(gis2)), scores, by = c('bin_id1', 'bin_id2'))

    # - Replace interactions, resolution and scores
    x@resolution <- bin.size
    x@interactions <- gis2
    x@scores <- S4Vectors::SimpleList(as.list(scores[, score_names]))

    return(x)
}

#' @title Contact map compartments
#' @name getCompartments
#' @rdname getCompartments
#' @description
#' 
#' Computes eigen vectors for each chromosome using cis contacts and extract 
#' chromosome compartments. 
#'
#' @param x A `HiCExperiment` object over a full genome
#' @param resolution Which resolution to use to compute eigen vectors
#' @param genome a BSgenome of DNAStringSet object associated with the Hi-C 
#'   contact matrix. 
#' @param chromosomes character or integer vector indicating which 
#' @param neigens Numver of eigen vectors to extract
#' @param sort_eigens Can be FALSE or one of c('Spearman', 'Pearson')
#' @param BPPARAM BiocParallel parallelization settings
#' @return A `HiCExperiment` object with additional `eigens` metadata containing the
#' normalized eigenvectors and a new "compartments" topologicalFeatures 
#' storing A and B compartments as a GRanges object. 
#'
#' @importFrom BiocParallel bplapply
#' @importFrom stats cor
#' @export
#' @examples 
#' library(HiContacts)
#' full_contacts_yeast <- contacts_yeast(full = TRUE)
#' comps <- getCompartments(full_contacts_yeast)
#' metadata(comps)$eigens

getCompartments <- function(
    x, 
    resolution = NULL, 
    genome = NULL, 
    chromosomes = NULL, 
    neigens = 3, 
    sort_eigens = FALSE, 
    BPPARAM = BiocParallel::bpparam()
) {
    message( "Going through preflight checklist..." )
    # - Check resolutions and chromosome subset
    if (is.null(resolution)) resolution <- resolution(x)
    resolution <- as.integer(resolution)
    chrs <- GenomicRanges::seqnames(GenomeInfoDb::seqinfo(x))
    names(chrs) <- chrs
    if (!is.null(chromosomes)) chrs <- chrs[chromosomes]

    # - If HiCExperiment comes from HiC-Pro object, make sure that the 
    #   imported matrix is genome-wide, and that `balanced` scores are available
    if ('regions' %in% names(metadata(x))) {
        if (!is.null(focus(x))) {
            message('Re-importing genome-wide HiCExperiment from HiC-Pro data files.')
            x <- HiCExperiment::HiCExperiment(
                fileName(x), bed = metadata(x)$regions
            )
        }
        if (!"balanced" %in% names(x)) {
            message('Balancing genome-wide matrix.')
            x <- normalize(x)
            scores(x, 'balanced') <- scores(x, 'ICE') 
        }
    }

    ## -- Parse contact matrix for each chromosome 
    message( "Parsing intra-chromosomal contacts for each chromosome..." )
    if (interactive()) {
        bpparam <- BiocParallel::SerialParam(progressbar = TRUE)
    }
    else {
        bpparam <- BiocParallel::SerialParam(progressbar = FALSE)
    }
    l_subs <- BiocParallel::bplapply(
        BPPARAM = bpparam, 
        chrs, 
        function(chr) {
            if ('regions' %in% names(metadata(x))) {
                x[chr]
            }
            else {
                HiCExperiment::HiCExperiment(
                    fileName(x), resolution = resolution, focus = chr
                )
            }
        }
    )
    names(l_subs) <- chrs

    ## -- Compute eigens on each chr. separately
    message( "Computing eigenvectors for each chromosome..." )
    compts <- BiocParallel::bplapply(
        BPPARAM = BPPARAM, 
        l_subs, 
        function(x_chr) {
            if (length(HiCExperiment::regions(x_chr)) <= neigens) {
                gr <- HiCExperiment::regions(x_chr)
                if (length(gr)) {
                    for (k in seq_len(neigens)) {
                        GenomicRanges::mcols(gr)[[paste0('E', k)]] <- 0
                    }
                    gr$eigen <- 0
                    gr$GC <- 0
                }
                else {
                    for (k in seq_len(neigens)) {
                        GenomicRanges::mcols(gr)[[paste0('E', k)]] <- NULL
                    }
                    gr$eigen <- NULL
                    gr$GC <- NULL
                }
                return(gr)
            } else {
                .eigChr(
                    x_chr, 
                    genome, 
                    neigens, 
                    sort_eigens = sort_eigens,
                    autocorrelation = FALSE,
                    clip_percentile = 0.999, 
                    ignore_diags = 2
                )
            }
        }
    ) |> GenomicRanges::GRangesList() |> unlist()

    ## -- Return initial HiCExperiment object with eigens in metadata
    A <- GenomicRanges::reduce(compts[compts$eigen > 0])
    A$compartment <- 'A'
    B <- GenomicRanges::reduce(compts[compts$eigen < 0])
    B$compartment <- 'B'
    cpts <- sort(c(A, B))
    HiCExperiment::topologicalFeatures(x, 'compartments') <- cpts
    metadata(x)$eigens <- compts
    return(x)

}

#' @importFrom RSpectra eigs_sym

.eigChr <- function(
    x_chr, 
    genome = NULL, 
    neigens = 3, 
    sort_eigens = FALSE,
    autocorrelation = FALSE, 
    clip_percentile = 0.99, 
    ignore_diags = 2
) {

    ## -- Get dense detrended matrix (with non-log2-detrended scores)
    if (!autocorrelation) { ## THIS IS THE DEFAULT (NO CORRELATION MATRIX)
        dx <- detrend(x_chr)
        gis <- InteractionSet::interactions(dx)
        gis$score <- HiCExperiment::scores(dx, 'detrended')
        gr <- HiCExperiment::regions(gis)
        m <- HiCExperiment::cm2matrix(HiCExperiment::gi2cm(gis))
        m <- 2^m ## Return to linear scale
        m <- m - 1 ## -- Subtract 1
    }
    else {
        dx <- autocorrelate(x_chr, ignore_ndiags = ignore_diags)
        gis <- InteractionSet::interactions(dx)
        gis$score <- HiCExperiment::scores(dx, 'autocorrelated')
        gr <- HiCExperiment::regions(gis)
        m <- HiCExperiment::cm2matrix(HiCExperiment::gi2cm(gis))
    }
    
    ## -- Remove ignored diagonals and NAs
    for (K in seq(-ignore_diags, ignore_diags, by = 1)) {
        sdiag(m, K) <- 0
    }
    m[is.na(m)] <- 0
    if (all(m == 0)) stop("Autocorrelation matrix is not complex enough. Try with a finer resolution.")

    ## -- Mask white lines from matrix
    m <- as.matrix(m)
    unmasked_bins <- rowSums(m) != 0
    m_no0 <- m[unmasked_bins, unmasked_bins]

    ## -- Clamp values at extremes
    m_no0 <- scales::oob_squish(
        m_no0, 
        stats::quantile(m_no0, probs = c(
            1-clip_percentile, clip_percentile
        ), na.rm = TRUE)
    )

    ## -- Get eigen values and vectors
    if (!requireNamespace("RSpectra", quietly = TRUE)) {
        message("RSpectra package not found. Falling back to base eigenvectors computation.")
        message("Install RSpectra package to perform faster eigenvectors computation on larger matrices.")
        message("install.packages('RSpectra')")
        eigen <- base::eigen(m_no0, symmetric = TRUE)
    } 
    else {
        eigen <- RSpectra::eigs_sym(m_no0, k = neigens)
    }
    
    ## -- Normalize eigenvector
    eigs <- eigen$vectors
    eigs <- eigs / apply(eigs, 2, function(x) as.numeric(sqrt(sum(x * x))))
    eigs <- lapply(seq_along(eigen$values), function(K) {
        eigs[, K] * as.numeric(sqrt(abs(eigen$values[K])))
    }) 
    names(eigs) <- paste0('E', seq_along(eigen$values))
    eigs <- dplyr::bind_cols(eigs)
    
    ## -- Recover full length eigenvector
    eigs_final <- as.data.frame(
        matrix(nrow = nrow(m), ncol = neigens, data = 0)
    )
    for (k in seq_len(neigens)) {
        eigs_final[unmasked_bins, k] <- eigs[, k]
        GenomicRanges::mcols(gr)[, paste0('E', k)] <- eigs_final[, k]
    }

    if (!is.null(genome)) {
        if (is(genome, "BSgenome") | is(genome, "DNAStringSet")) {
            ## -- Get GC content
            gr$phasing <- as.numeric(Biostrings::letterFrequency(
                Biostrings::getSeq(genome, gr), letters = 'GC') / 
                GenomicRanges::width(gr))
        }
        else if (is(genome, "TxDb")) {
            ## -- Get gene density
            if (!requireNamespace("GenomicFeatures", quietly = TRUE)) {
                message("Install GenomicFeatures package to phase eigenvector using a `TxDb` object.")
                message("install.packages('GenomicFeatures')")
                gr$phasing <- 1
            } 
            else {
                cov <- GenomicFeatures::genes(genome) |> 
                    GenomicRanges::coverage()
                gr$phasing <- BiocGenerics::mean(cov[gr])
            }
        }
        else if (is(genome, "RleList")) {
            ## -- Get average coverage per bin
            gr$phasing <- BiocGenerics::mean(cov[gr])
        }
        else {
            stop("`genome` argument not recognized.")
        }
        ## -- Get matching eigenvector and phase it with GC
        gr <- .eigGCPhasing(gr, neigens, unmasked_bins, sort_eigens)
    }
    else {
        message("Caution! No genome is provided. The first eigenvector has ", 
        "been selected and no phasing has been performed.")
        gr$eigen <- gr$E1
    }

    return(gr)

}

.eigGCPhasing <- function(gr, neigens, unmasked_bins, sort_eigens) {
    if (isFALSE(sort_eigens)) {
        method <- 'spearman'
    }
    else if (sort_eigens == 'Spearman') {
        method <- 'spearman'
    }
    else if (sort_eigens == 'Pearson') {
        method <- 'pearson'
    }
    else {
        stop("`sort_eigens` can be FALSE or one of 'Spearman' or 'Pearson'")
    }
    cors <- lapply(seq_len(neigens), function(K) {
        vec <- GenomicRanges::mcols(gr)[, paste0('E', K)]
        stats::cor(
            gr$phasing[unmasked_bins], 
            vec[unmasked_bins], 
            method = method
        )
    }) |> unlist()

    # -- Make all eigens positively correlate with phasing
    for (i in seq_len(neigens)) {
        if (cors[i] < 0) {
            GenomicRanges::mcols(gr)[, paste0('E', i)] <- -GenomicRanges::mcols(gr)[, paste0('E', i)]
            cors[i] <- -cors[i]
        }
    }

    # -- Sort eigens by highest corr.
    if (!isFALSE(sort_eigens)) {
        gr0 <- gr
        cors0 <- cors
        o <- order(cors, decreasing = TRUE)
        for (i in seq_len(neigens)) {
            GenomicRanges::mcols(gr0)[, paste0('E', i)] <- GenomicRanges::mcols(gr)[, paste0('E', o[i])]
            cors0[i] <- cors[o[i]]
        }
        gr <- gr0
        cors <- cors0
        gr$eigen <- gr$E1
    }
    else {
        gr$eigen <- gr$E1
    }
    
    # -- Print correlations
    fcors <- paste(round(cors, 2), collapse = ' / ')
    message(paste0(
        "seqnames `", 
        GenomicRanges::seqnames(gr)[1], "` : ", fcors
    ))

    return(gr)
}

#' Compute the law of distance-dependent contact frequency, a.k.a. P(s)
#' 
#' P(s) will be approximated if no pairs are provided, or the exact P(s) 
#' will be computed if a `.pairs` file is added to the `HiCExperiment` object 
#' using `pairsFile(x) <- "..."`. 
#' 
#' @name Ps
#' @aliases distanceLaw,GInteractions,missing-method
#' @aliases distanceLaw,HiCExperiment,missing-method
#' @aliases distanceLaw,PairsFile,missing-method
#' @aliases distanceLaw,PairsFile,GRanges-method
#' @aliases localDistanceLaw
#' 
#' @param x A `HiCExperiment` object
#' @param by_chr by_chr
#' @param filtered_chr filtered_chr
#' @param chunk_size For pairs files which do not fit in memory, pick a number 
#' of pairs to parse by chunks (1e7 should be a good compromise)
#' @param coords GRanges to specify which genomic loci to use when computing P(s)
#' @param ... Arguments passed to corresponding method
#' @return a tibble
#'
#' @import tibble
#' @importFrom dplyr filter
#' @importFrom dplyr mutate
#' @importFrom dplyr group_by
#' @importFrom dplyr summarize
#' @importFrom dplyr left_join
#' @importFrom dplyr group_split
#' @importFrom dplyr slice
#' @importFrom dplyr pull
#' @importFrom dplyr bind_rows
#' @importFrom dplyr select
#' @importFrom dplyr tally
#' @importFrom dplyr arrange
#' @importFrom dplyr lead
#' @importFrom dplyr cur_data
#' @importFrom BiocIO import
#' @importFrom BiocGenerics path
#' @importFrom readr read_tsv_chunked
#' @importFrom readr DataFrameCallback
#' @examples 
#' contacts_yeast <- contacts_yeast()
#' ps <- distanceLaw(contacts_yeast)
#' ps
#' local_ps <- localDistanceLaw(
#'     contacts_yeast,
#'     GenomicRanges::GRanges(
#'         c("telomere" = "II:1-20000", "arm" = "II:300001-700000")
#'     )
#' )
#' local_ps
NULL

#' @rdname Ps
#' @export 

setMethod("distanceLaw", signature(x = "GInteractions", coords = "missing"), function(
    x, 
    by_chr = FALSE 
) {
    df <- tibble::tibble(
        chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(x)[['first']])),
        distance = InteractionSet::pairdist(x, type = 'gap')
    ) |> 
        tidyr::drop_na() |> 
        dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
    if (by_chr) {
        df <- dplyr::group_by(df, chr, binned_distance)
    } 
    else {
        df <- dplyr::group_by(df, binned_distance)
    }
    d <- dplyr::tally(df, name = 'ninter') |>
        dplyr::mutate(p = ninter/sum(ninter)) |> 
        dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
        dplyr::mutate(norm_p = p / binwidth)
    if (by_chr) {
        d <- dplyr::group_by(d, chr)
    } 
    else {
        d <- d
    }
    if (by_chr) {
        d <- dplyr::group_by(d, chr)
    } 
    else {
        d <- d
    }
    ps <- dplyr::group_split(d) |> 
        lapply(function(x) {
            dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
            dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
            dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::pick(slope)))))
        }) |> 
        dplyr::bind_rows()
    if (by_chr) {
        ps <- dplyr::select(ps, chr, binned_distance, p, norm_p, norm_p_unity, slope) |> 
            dplyr::arrange(binned_distance)
    } 
    else {
        ps <- dplyr::select(ps, binned_distance, p, norm_p, norm_p_unity, slope) |> 
            dplyr::arrange(binned_distance)
    }
    return(ps)
})

#' @rdname Ps
#' @export 

setMethod("distanceLaw", signature(x = "HiCExperiment", coords = "missing"), function(
    x, 
    by_chr = FALSE, 
    filtered_chr = c('XII', 'chrXII', 'chr12', '12', 'Mito', 'MT', 'chrM')
) {
    pairsFile <- HiCExperiment::pairsFile(x)
    if (!is.null(pairsFile)) {
        distanceLaw(
            PairsFile(pairsFile), by_chr = by_chr, filtered_chr = filtered_chr
        )
    }
    else {
        # stop("Please provide a pairsFile for `x`. Aborting now.")
        message("pairsFile not specified. The P(s) curve will be an approximation.")
        pairs <- InteractionSet::interactions(x)
        pairs$score <- HiCExperiment::scores(x, 'count')
        df <- tibble::tibble(
            chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(pairs)[['first']])),
            distance = InteractionSet::pairdist(pairs, type = 'gap'),
            n = pairs$score
        ) |> 
            tidyr::drop_na() |> 
            dplyr::filter(!chr %in% filtered_chr) |> 
            dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
        if (by_chr) {
            df <- dplyr::group_by(df, chr, binned_distance)
        } 
        else {
            df <- dplyr::group_by(df, binned_distance)
        }
        d <- dplyr::summarize(df, ninter = sum(n)) |>
            dplyr::mutate(p = ninter/sum(ninter)) |> 
            dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
            dplyr::mutate(norm_p = p / binwidth)
        if (by_chr) {
            d <- dplyr::group_by(d, chr)
        } 
        else {
            d <- d
        }
        if (by_chr) {
            d <- dplyr::group_by(d, chr)
        } 
        else {
            d <- d
        }
        ps <- dplyr::group_split(d) |> 
            lapply(function(x) {
                dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
                dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
                dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::pick(slope)))))
            }) |> 
            dplyr::bind_rows()
        if (by_chr) {
            ps <- dplyr::select(ps, chr, binned_distance, p, norm_p, norm_p_unity, slope) |> 
                dplyr::arrange(binned_distance)
        } 
        else {
            ps <- dplyr::select(ps, binned_distance, p, norm_p, norm_p_unity, slope) |> 
                dplyr::arrange(binned_distance)
        }
        return(ps)
    }
})

#' @rdname Ps
#' @export 

setMethod("distanceLaw", signature(x = "PairsFile", coords = "missing"), function(
    x, 
    by_chr = FALSE, 
    filtered_chr = c('XII', 'chrXII', 'chr12', '12', 'Mito', 'MT', 'chrM'), 
    chunk_size = 100000
) {
    message("Importing pairs file ", BiocGenerics::path(x), " in memory. This may take a while...")
    if (chunk_size == 0) { ### Fit everything in memory
        pairs <- BiocIO::import(x)
        df <- tibble::tibble(
            chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(pairs)[['first']])),
            distance = pairs$distance
        ) |> 
            tidyr::drop_na() 
    }
    else {
        f <- function(x, pos) {
            subx <- x[x[, 2] == x[, 4], ]
            tibble::tibble(
                chr = subx[[2]], 
                distance = subx[[5]] - subx[[3]]
            )
        }
        df <- readr::read_tsv_chunked(
            file = BiocGenerics::path(x), 
            callback = readr::DataFrameCallback$new(f), 
            chunk_size = chunk_size, 
            col_names = FALSE, 
            comment = "#", 
            show_col_types = FALSE
        ) 
    }
    df <- df |> 
        dplyr::filter(!chr %in% filtered_chr) |> 
        dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
    if (by_chr) {
        df <- dplyr::group_by(df, chr, binned_distance)
    } 
    else {
        df <- dplyr::group_by(df, binned_distance)
    }
    d <- dplyr::tally(df, name = 'ninter') |>
        dplyr::mutate(p = ninter/sum(ninter)) |> 
        dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
        dplyr::mutate(norm_p = p / binwidth)
    if (by_chr) {
        d <- dplyr::group_by(d, chr)
    } 
    else {
        d <- d
    }
    if (by_chr) {
        d <- dplyr::group_by(d, chr)
    } 
    else {
        d <- d
    }
    ps <- dplyr::group_split(d) |> 
        lapply(function(x) {
            dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
            dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
            dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::pick(slope)))))
        }) |> 
        dplyr::bind_rows()
    if (by_chr) {
        ps <- dplyr::select(ps, chr, binned_distance, p, norm_p, norm_p_unity, slope) |> 
            dplyr::arrange(binned_distance)
    } 
    else {
        ps <- dplyr::select(ps, binned_distance, p, norm_p, norm_p_unity, slope) |> 
            dplyr::arrange(binned_distance)
    }
    return(ps)
})

#' @rdname Ps
#' @export 

setMethod("distanceLaw", signature(x = "HiCExperiment", coords = "GRanges"), function(
    x, 
    coords,
    chunk_size = 100000
) {
    pairsFile <- HiCExperiment::pairsFile(x)
    if (!is.null(pairsFile)) {
        distanceLaw(
            PairsFile(pairsFile), coords = coords, chunk_size = chunk_size
        )
    }
    else {
        # stop("Please provide a pairsFile for `x`. Aborting now.")
        message("pairsFile not specified. The P(s) curve will be an approximation.")
        pairs <- InteractionSet::interactions(x)
        pairs$score <- HiCExperiment::scores(x, 'count')
        pairs <- subsetByOverlaps(pairs, coords)
        df <- tibble::tibble(
            chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(pairs)[['first']])),
            distance = InteractionSet::pairdist(pairs, type = 'gap'),
            n = pairs$score
        ) |> 
            tidyr::drop_na() |> 
            dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
        df <- dplyr::group_by(df, binned_distance)
        d <- dplyr::summarize(df, ninter = sum(n)) |>
            dplyr::mutate(p = ninter/sum(ninter)) |> 
            dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
            dplyr::mutate(norm_p = p / binwidth)
        ps <- dplyr::group_split(d) |> 
            lapply(function(x) {
                dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
                dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
                dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::pick(slope)))))
            }) |> 
            dplyr::bind_rows()
        ps <- dplyr::select(ps, binned_distance, p, norm_p, norm_p_unity, slope) |> 
            dplyr::arrange(binned_distance)
        return(ps)
    }

})

#' @rdname Ps
#' @export 

setMethod("distanceLaw", signature(x = "PairsFile", coords = "GRanges"), function(
    x, 
    coords,
    chunk_size = 100000
) {
    message("Importing pairs file ", BiocGenerics::path(x), " in memory. This may take a while...")
    if (chunk_size == 0) { ### Fit everything in memory
        pairs <- BiocIO::import(x)
        pairs <- subsetByOverlaps(pairs, coords)
        df <- tibble::tibble(
            chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(pairs)[['first']])),
            distance = pairs$distance
        ) |> 
            tidyr::drop_na() 
    }
    else {
        f <- function(.x, pos) {
            pairs <- InteractionSet::GInteractions(
                GenomicRanges::GRanges(
                    .x[[2]],
                    IRanges::IRanges(.x[[3]], width = 1), 
                    strand = .x[[6]]
                ), 
                GenomicRanges::GRanges(
                    .x[[4]],
                    IRanges::IRanges(.x[[5]], width = 1), 
                    strand = .x[[7]]
                )
            )
            pairs <- subsetByOverlaps(pairs, coords)
            df <- tibble::tibble(
                chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(pairs)[['first']])),
                distance = pairdist(pairs)
            ) |> 
                tidyr::drop_na() 
        }
        df <- readr::read_tsv_chunked(
            file = BiocGenerics::path(x), 
            callback = readr::DataFrameCallback$new(f), 
            chunk_size = chunk_size, 
            col_names = FALSE, 
            comment = "#", 
            show_col_types = FALSE
        ) 
    }
    d <- dplyr::mutate(df, 
        binned_distance = PsBreaks()$break_pos[
            findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)
        ]
    ) |> dplyr::group_by(binned_distance) |> 
        dplyr::tally(name = 'ninter') |>
        dplyr::mutate(p = ninter/sum(ninter)) |> 
        dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
        dplyr::mutate(norm_p = p / binwidth)
    ps <- dplyr::group_split(d) |> 
        lapply(function(x) {
            dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
            dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
            dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::pick(slope)))))
        }) |> 
        dplyr::bind_rows()
    ps <- dplyr::select(ps, binned_distance, p, norm_p, norm_p_unity, slope) |> 
        dplyr::arrange(binned_distance)
    return(ps)
})

#' @rdname Ps
#' @export 

localDistanceLaw <- function(
    x, 
    coords = coords
) {
    `%over%` <- IRanges::`%over%`
    if(is.null(names(coords))) names(coords) <- as(coords, 'character')
    pairsFile <- HiCExperiment::pairsFile(x)
    if (is.null(pairsFile)) {
        message("pairsFile not specified. The P(s) curve will be an approximation.")
        an_ <- anchors(x)
        d <- lapply(seq_along(coords), function(K) {
            gr <- coords[K]
            sub <- an_[['first']] %over% gr | an_[['second']] %over% gr
            x <- x[sub]
            pairs <- interactions(x)
            pairs$score <- HiCExperiment::scores(x, 'count')
            df <- tibble::tibble(
                chr = as.vector(GenomeInfoDb::seqnames(an_[['first']][sub])),
                distance = InteractionSet::pairdist(pairs, type = 'gap'),
                n = pairs$score
            ) |> 
                tidyr::drop_na() |> 
                dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
            df <- dplyr::group_by(df, binned_distance)
            d <- dplyr::summarize(df, ninter = sum(n)) |>
                dplyr::mutate(p = ninter/sum(ninter)) |> 
                dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
                dplyr::mutate(norm_p = p / binwidth) |>
                dplyr::mutate(coords = names(coords)[K])
        }) |> dplyr::bind_rows()
    }
    else {
        message("Importing pairs file ", pairsFile, " in memory. This may take a while...")
        pairs <- BiocIO::import(pairsFile, format = 'pairs')
        an_ <- anchors(pairs)
        d <- lapply(seq_along(coords), function(K) {
            gr <- coords[K]
            sub <- an_[['first']] %over% gr | an_[['second']] %over% gr
            subpairs <- pairs[sub]
            df <- tibble::tibble(
                chr = as.vector(GenomeInfoDb::seqnames(InteractionSet::anchors(subpairs)[['first']])),
                distance = subpairs$distance
            ) |> 
                tidyr::drop_na() |> 
                dplyr::mutate(binned_distance = PsBreaks()$break_pos[findInterval(distance, vec = PsBreaks()$break_pos, all.inside = TRUE)])
            df <- dplyr::group_by(df, binned_distance)
            dplyr::tally(df, name = 'ninter') |>
                dplyr::mutate(p = ninter/sum(ninter)) |> 
                dplyr::left_join(PsBreaks(), by = c('binned_distance' = 'break_pos')) |> 
                dplyr::mutate(norm_p = p / binwidth) |>
                dplyr::mutate(coords = names(coords)[K])
        }) |> dplyr::bind_rows()
    }
    d <- dplyr::group_by(d, coords)
    ps <- dplyr::group_split(d) |> 
        lapply(function(x) {
            dplyr::mutate(x, norm_p_unity = norm_p / {dplyr::slice(x, which.min(abs(x$binned_distance - 100000))) |> dplyr::pull(norm_p)}) |> 
            dplyr::mutate(slope = (log10(dplyr::lead(norm_p)) - log10(norm_p)) / (log10(dplyr::lead(binned_distance)) - log10(binned_distance))) |> 
            dplyr::mutate(slope = c(0, predict(loess(slope ~ binned_distance, span = 0.5, data = dplyr::cur_data()))))
        }) |> 
        dplyr::bind_rows()
    ps <- dplyr::select(ps, coords, binned_distance, p, norm_p, norm_p_unity, slope) |> 
        dplyr::arrange(binned_distance)
    return(ps)
}

PsBreaks <- function() {
    structure(list(break_pos = c(
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
        12, 13, 14, 16, 17, 19, 21, 23, 26, 28, 31, 34, 37, 41, 45, 50,
        55, 60, 66, 73, 80, 88, 97, 107, 117, 129, 142, 156, 172, 189,
        208, 229, 252, 277, 304, 335, 368, 405, 446, 490, 539, 593, 653,
        718, 790, 869, 956, 1051, 1156, 1272, 1399, 1539, 1693, 1862,
        2048, 2253, 2479, 2726, 2999, 3299, 3629, 3992, 4391, 4830, 5313,
        5844, 6429, 7072, 7779, 8557, 9412, 10354, 11389, 12528, 13781,
        15159, 16675, 18342, 20176, 22194, 24413, 26855, 29540, 32494,
        35743, 39318, 43249, 47574, 52332, 57565, 63322, 69654, 76619,
        84281, 92709, 101980, 112178, 123396, 135735, 149309, 164240,
        180664, 198730, 218603, 240463, 264510, 290961, 320057, 352063,
        387269, 425996, 468595, 515455, 567000, 623700, 686070, 754677,
        830145, 913160, 1004475, 1104923, 1215415, 1336957, 1470653,
        1617718, 1779490, 1957439, 2153182, 2368501, 2605351, 2865886,
        3152474, 3467722, 3814494, 4195943, 4615538, 5077092, 5584801,
        6143281, 6757609, 7433370, 8176707, 8994377, 9893815, 10883197,
        11971516, 13168668, 14485535, 15934088, 17527497, 19280247, 21208271,
        23329099, 25662008, 28228209, 31051030, 34156133, 37571747, 41328921,
        45461813, 50007995, 55008794, 60509674, 66560641, 73216705, 80538375,
        88592213, 97451434, 107196578, 117916236, 129707859, 142678645,
        156946509, 172641160, 189905276, 208895804, 229785385, 252763923,
        278040315, 305844347, 336428781, 370071660, 407078826, 447786708,
        492565379, 541821917, 596004109, 655604519, 721164971, 793281468,
        872609615, 959870577, 1055857635, 1161443398
    ), binwidth = c(
        1,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 2, 2, 3, 2, 3, 3,
        3, 4, 4, 5, 5, 5, 6, 7, 7, 8, 9, 10, 10, 12, 13, 14, 16, 17,
        19, 21, 23, 25, 27, 31, 33, 37, 41, 44, 49, 54, 60, 65, 72, 79,
        87, 95, 105, 116, 127, 140, 154, 169, 186, 205, 226, 247, 273,
        300, 330, 363, 399, 439, 483, 531, 585, 643, 707, 778, 855, 942,
        1035, 1139, 1253, 1378, 1516, 1667, 1834, 2018, 2219, 2442, 2685,
        2954, 3249, 3575, 3931, 4325, 4758, 5233, 5757, 6332, 6965, 7662,
        8428, 9271, 10198, 11218, 12339, 13574, 14931, 16424, 18066,
        19873, 21860, 24047, 26451, 29096, 32006, 35206, 38727, 42599,
        46860, 51545, 56700, 62370, 68607, 75468, 83015, 91315, 100448,
        110492, 121542, 133696, 147065, 161772, 177949, 195743, 215319,
        236850, 260535, 286588, 315248, 346772, 381449, 419595, 461554,
        507709, 558480, 614328, 675761, 743337, 817670, 899438, 989382,
        1088319, 1197152, 1316867, 1448553, 1593409, 1752750, 1928024,
        2120828, 2332909, 2566201, 2822821, 3105103, 3415614, 3757174,
        4132892, 4546182, 5000799, 5500880, 6050967, 6656064, 7321670,
        8053838, 8859221, 9745144, 10719658, 11791623, 12970786, 14267864,
        15694651, 17264116, 18990528, 20889581, 22978538, 25276392, 27804032,
        30584434, 33642879, 37007166, 40707882, 44778671, 49256538, 54182192,
        59600410, 65560452, 72116497, 79328147, 87260962, 95987058, 105585763,
        116144340
    )), row.names = c(NA, -205L), class = c(
        "tbl_df", "tbl",
        "data.frame"
    ))
}

#' HiContacts plotting functionalities
#' 
#' @name HiContacts-plots
#' 
#' @description 
#' Several plots can be generated in HiContacts: 
#' - Hi-C contact matrices
#' - Distance-dependant interaction frequency decay (a.k.a. "Distance law" or "P(s)")
#' - Virtual 4C profiles
#' - Scalograms
#' - Saddle plots
NULL 

################################################################################
#                                                                              #
#                                 Matrix                                       #
#                                                                              #
################################################################################

#' Plotting a contact matrix
#'
#' @name plotMatrix
#' @aliases plotMatrix,HiCExperiment-method
#' @aliases plotMatrix,HiCExperiment-method
#' @aliases plotMatrix,GInteractions-method
#' @aliases plotMatrix,matrix-method
#' 
#' @param x A HiCExperiment object
#' @param compare.to Compare to a second HiC matrix in the lower left corner
#' @param use.scores Which scores to use in the heatmap
#' @param scale Any of 'log10', 'log2', 'linear', 'exp0.2' (Default: 'log10')
#' @param limits color map limits
#' @param maxDistance maximum distance. If provided, the heatmap is plotted 
#'   horizontally
#' @param loops Loops to plot on top of the heatmap, provided as `GInteractions`
#' @param borders Borders to plot on top of the heatmap, provided as `GRanges`
#' @param tracks Named list of bigwig tracks imported as `Rle`
#' @param dpi DPI to create the plot (Default: 500)
#' @param rasterize Whether the generated heatmap is rasterized or vectorized 
#' (Default: TRUE)
#' @param symmetrical Whether to enforce a symetrical heatmap (Default: TRUE)
#' @param chrom_lines Whether to display separating lines between chromosomes, 
#' should any be necessary (Default: TRUE)
#' @param show_grid Whether to display an underlying grid (Default: FALSE)
#' @param caption Whether to display a caption (Default: TRUE)
#' @param cmap Color scale to use. (Default: bgrColors() if limits are c(-1, 1)
#' and coolerColors() otherwise)
#' @return ggplot object
#'
#' @importFrom ggrastr geom_tile_rast
#' @importFrom tibble as_tibble
#' @importFrom dplyr mutate
#' @importFrom dplyr select
#' @importFrom dplyr group_by
#' @importFrom dplyr summarize
#' @importFrom dplyr slice_max
#' @importFrom dplyr distinct
#' @importFrom dplyr left_join
#' @importFrom dplyr rename
#' @importFrom dplyr left_join
#' @importFrom dplyr rename
#' @importFrom GenomicRanges seqnames
#' @importFrom scales oob_squish
#' @importFrom scales unit_format
#' @examples 
#' contacts_yeast <- contacts_yeast()
#' plotMatrix(
#'     contacts_yeast, 
#'     use.scores = 'balanced', 
#'     scale = 'log10', 
#'     limits = c(-4, -1)
#' )
NULL 

#' @rdname plotMatrix
#' @export

setMethod("plotMatrix", "HiCExperiment", function(
    x, 
    compare.to = NULL, 
    use.scores = 'balanced', 
    scale = 'log10', 
    maxDistance = NULL, 
    loops = NULL, 
    borders = NULL, 
    tracks = NULL, 
    limits = NULL, 
    dpi = 500, 
    rasterize = TRUE, 
    symmetrical = TRUE, 
    chrom_lines = TRUE, 
    show_grid = FALSE, 
    cmap = NULL, 
    caption = TRUE  
) {
    if (!use.scores %in% names(scores(x))) 
        stop(paste0("Queried scores not found."))

    if (!is.null(compare.to)) {
        gis_x <- interactions(x)
        gis_y <- interactions(compare.to)
        gis <- c(
            as(gis_y, 'GInteractions'), 
            InteractionSet::swapAnchors(
                as(gis_x, 'GInteractions'), 
                mode = 'reverse'
            )
        )
        gis <- gis[pairdist(gis) != 0]
        p <- plotMatrix(
            gis, 
            use.scores = use.scores, 
            scale = scale, 
            maxDistance = maxDistance, 
            loops = loops, 
            borders = borders, 
            tracks = tracks, 
            limits = limits, 
            dpi = dpi, 
            rasterize = rasterize, 
            symmetrical = FALSE, 
            chrom_lines = chrom_lines, 
            show_grid = show_grid, 
            cmap = cmap  
        )
    }
    else {
        p <- plotMatrix(
            interactions(x), 
            use.scores = use.scores, 
            scale = scale, 
            maxDistance = maxDistance, 
            loops = loops, 
            borders = borders, 
            tracks = tracks, 
            limits = limits, 
            dpi = dpi, 
            rasterize = rasterize, 
            symmetrical = symmetrical, 
            chrom_lines = chrom_lines, 
            show_grid = show_grid, 
            cmap = cmap  
        )
    }
    if (caption & is.null(tracks)) {
        f <- fileName(x)
        if (!is.null(compare.to)) {
            f <- paste0(fileName(x), ' | ', fileName(compare.to))
            print(f)
        }
        p <- p + ggplot2::labs(
            caption = paste(
                sep = '\n',
                paste0('file: ', f), 
                paste0('resolution: ', resolution(x)), 
                paste0('focus: ', focus(x)),
                paste0('scores: ', use.scores),
                paste0('scale: ', scale)
            )
        )
    }
    return(p)
})

#' @rdname plotMatrix
#' @export

setMethod("plotMatrix", "GInteractions", function(
    x, 
    use.scores = NULL, 
    scale = 'log10', 
    maxDistance = NULL, 
    loops = NULL, 
    borders = NULL, 
    tracks = NULL, 
    limits = NULL, 
    dpi = 500, 
    rasterize = TRUE, 
    symmetrical = TRUE, 
    chrom_lines = TRUE, 
    show_grid = FALSE, 
    cmap = NULL  
) {
    `%over%` <- IRanges::`%over%`

    ## -- Switch `symmetrical` to FALSE if only trans interactions
    gis <- x
    onlyTRANS <- !any(S4Vectors::first(gis) == S4Vectors::second(gis))
    if (onlyTRANS & symmetrical) symmetrical <- FALSE

    ## -- Extract scores
    if (!is.null(use.scores)) {
        gis$score <- S4Vectors::mcols(gis)[, use.scores]
    }
    else {
        if ("balanced" %in% colnames(S4Vectors::mcols(gis))) {
            gis$score <- S4Vectors::mcols(gis)[, 'balanced']
        } 
        else {
            gis$score <- S4Vectors::mcols(gis)[, 'count']
        }
    }

    ## -- Put metric to plot in `score` column
    has_negative_scores <- any(gis$score < 0, na.rm = TRUE)

    ## -- Scale score
    if (scale == 'log10') {
        gis$score <- log10(gis$score)
    } 
    else if (scale == 'log2') {
        gis$score <- log2(gis$score)
    }
    else if (scale == 'exp0.2') {
        gis$score <- gis$score^0.2
    }
    else if (scale == 'linear') {
        gis$score <- gis$score
    }

    ## -- Set matrix limits
    if (!is.null(limits)) {
        m <- limits[1]
        M <- limits[2]
    }
    else {
        if (onlyTRANS) {
            .scores <- gis$score
        } 
        else {
            .scores <- gis$score[pairdist(gis) != 0 & !is.na(pairdist(gis) != 0)]
        }
        .scores <- .scores[!is.na(.scores)]
        .scores <- .scores[!is.infinite(.scores)]
        M <- max(.scores)
        m <- min(.scores)
        limits <- c(m, M)
    }

    ## -- Choose color map 
    if (is.null(cmap)) {
        if (has_negative_scores | {m == -1 & M == 1}) {
            cmap <- bgrColors()
        }
        else {
            cmap <- coolerColors()
        }
    }
    
    # -- Define plotting approach
    if (rasterize) {
        plotFun <- ggrastr::geom_tile_rast(
            raster.dpi = dpi,
            width = GenomicRanges::width(gis)[[1]][1], 
            height = GenomicRanges::width(gis)[[1]][1]
        )
    }
    else {
        plotFun <- ggplot2::geom_tile()
    }

    ## -- If loops are provided, filter them and add
    if (!is.null(loops) & is.null(maxDistance)) {
        filtered_loops <- tibble::as_tibble(
            loops[anchors(loops)[['first']] %over% gis & anchors(loops)[['second']] %over% gis]
        )
        p_loops <- ggplot2::geom_point(
            data = filtered_loops, 
            mapping = ggplot2::aes(
                x = start1+(end1-start1)/2, y = start2+(end2-start2)/2
            ), 
            inherit.aes = FALSE, 
            shape = 21, 
            fill = "#76eaff00", 
            color = "#000000", 
            size = 3
        )
    }
    else {
        p_loops <- NULL
    }

    ## -- If borders are provided, filter them and add
    if (!is.null(borders) & is.null(maxDistance)) {
        filtered_borders <- tibble::as_tibble(
            borders[borders %over% gis]
        )
        p_borders <- ggplot2::geom_point(
            data = filtered_borders, 
            mapping = ggplot2::aes(x = start+(end-start)/2, y = start+(end-start)/2),
            inherit.aes = FALSE, 
            shape = 23, 
            fill = "#76eaff", 
            color = "#000000", 
            size = 1
        )
    }
    else {
        p_borders <- NULL
    }

    ## -- If tracks are provided, filter them and add
    if (!is.null(tracks) & is.null(maxDistance)) {
        re <- reduce(regions(gis))
        # if (length(re) > 1) {
        #     warning("Input track cannot be aligned to interactions with non-contiguous regions.")
        #     p_tracks <- ggplot2::ggplot() + ggplot2::theme_void()
        # }
        # else {
            ntracks <- length(tracks)
            step <- GenomicRanges::width(InteractionSet::anchors(gis[1])[[1]])/5
            filtered_tracks <- purrr::imap(tracks, ~ .x[re] |> 
                tibble::as_tibble() |> 
                dplyr::group_by(group=(0:(n()-1))%/%{step}) |> 
                dplyr::summarize(value = mean(value, na.rm = TRUE)) |> 
                dplyr::mutate(
                    # pos = seq(IRanges::start(re), IRanges::end(re), by = step) + step/2, 
                    track = .y
                )
            ) |> dplyr::bind_rows(.id = 'track')
            p_tracks <- ggplot2::ggplot() + 
                ggplot2::geom_area(
                    data = filtered_tracks, 
                    mapping = ggplot2::aes(x = group, y = value),
                    fill = "#393939", 
                    color = NA, 
                ) + 
                ggplot2::scale_x_continuous(expand = c(0, 0)) +
                ggplot2::scale_y_reverse(expand = c(0, 0)) +
                ggplot2::facet_grid(track~., scales = 'free') + 
                ggplot2::theme_void() + 
                ggplot2::theme(
                    strip.background = ggplot2::element_blank(), 
                    strip.text = ggplot2::element_blank()
                )
        # }
    }
    else {
        p_tracks <- ggplot2::ggplot() + ggplot2::theme_void()
    }

    # -- Check number of chromosomes that were extracted
    nseqnames <- length(unique(as.vector(
        GenomicRanges::seqnames(InteractionSet::regions(gis))
    )))

    if (nseqnames == 1 | {nseqnames == 2 & all(
        GenomicRanges::seqnames(InteractionSet::anchors(gis, "first")) != 
        GenomicRanges::seqnames(InteractionSet::anchors(gis, "second"))
    )}) { ## SINGLE CHROMOSOME MAP or 2 CHROMOSOMES TRANS INTERSECTION
        
        if (is.null(maxDistance)) { ##### REGULAR SQUARE/RECTANGULAR (IF TRANS ONLY) MATRIX
            ## -- Convert gis to table and extract x/y
            mat <- gis |>
                tibble::as_tibble() |>
                dplyr::mutate(
                    x = floor(end1 - (end1 - start1) / 2),
                    y = floor(end2 - (end2 - start2) / 2)
                ) |> 
                tidyr::drop_na(score)

            ## -- Clamp scores to limits
            mat$score = scales::oob_squish(mat$score, c(m, M))

            ## -- Add lower triangular matrix scores (if symetrical)
            if (symmetrical) {
                mat <- rbind(
                    mat, 
                    mat |> 
                        dplyr::mutate(x2 = y, y = x, x = x2) |> 
                        dplyr::select(-x2)
                )
                bb <- InteractionSet::boundingBox(x)
                bbOne <- InteractionSet::anchors(bb)[[1]]
                bbTwo <- InteractionSet::anchors(bb)[[2]]
                fraction_ov <- GenomicRanges::width(GenomicRanges::pintersect(bbOne, bbTwo)) /
                    {{GenomicRanges::width(bbOne) + GenomicRanges::width(bbTwo)}/2}
                if (fraction_ov > 0.9) {
                    coords <- c(bbOne, bbTwo)
                    mat <- mat |> 
                        dplyr::filter(x >= GenomicRanges::start(coords[2]) & 
                            x <= GenomicRanges::end(coords[2])) |> 
                        dplyr::filter(y >= GenomicRanges::start(coords[1]) & 
                            y <= GenomicRanges::end(coords[1]))
                }
            } 

            ## -- Plot matrix
            p <- ggMatrix(mat, cols = cmap, limits = limits, grid = show_grid) +
                plotFun +
                p_loops + 
                p_borders + 
                ggplot2::labs(
                    x = unique(mat$seqnames1),
                    y = "Genome coordinates"
                ) + 
                ggplot2::scale_x_continuous(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6), position = 'top') +
                ggplot2::scale_y_reverse(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6))

            ## -- Add tracks 
            if (length(p_tracks$layers)) {
                p <- p + 
                ggplot2::theme(
                    plot.margin = grid::unit(c(5.5, 5.5, 5.5 + 10*ntracks, 5.5), "points")
                ) + patchwork::inset_element(
                    p_tracks, 
                    left = 0, 
                    bottom = -0.1*ntracks, 
                    right = 1, 
                    top = 0, 
                    align_to = 'panel', 
                    clip = FALSE
                ) 
            }
            
        }
        
        else { ##### HORIZONTAL TRIANGULAR MATRIX
            ## -- Convert gis to table and extract x/y
            df <- gis |>
                tibble::as_tibble() |>
                dplyr::mutate(
                    diag = center2 - center1, 
                    x = center1 + (center2 - center1)/2
                ) |> 
                tidyr::drop_na(score) |> 
                dplyr::filter(diag <= maxDistance)

            ## -- Clamp scores to limits
            df <- dplyr::mutate(df, score = scales::oob_squish(score, c(m, M)))

            ## -- Plot matrix
            p <- ggHorizontalMatrix(df, cols = cmap, limits = limits) +
                plotFun +
                p_loops + 
                p_borders + 
                ggplot2::labs(
                    x = "Genomic location",
                    y = "Distance"
                )
        }
    }

    else { ##### MULTI-CHROMOSOMES MAP

        ## -- Abort if more than 1 chr. is plotted AND maxDistance is provided
        if (!is.null(maxDistance)) stop(
            "Horizontal, multi-chromosomes maps are not supported."
        )

        chroms <- gis |>
            tidyr::as_tibble() |>
            dplyr::group_by(seqnames2) |>
            dplyr::slice_max(order_by = end2, n = 1) |>
            dplyr::select(seqnames2, end2) |>
            dplyr::distinct()
        chroms$cumlength <- cumsum(c(0, chroms$end2)[seq_len(nrow(chroms))])
        chroms$end2 <- NULL
        mat <- gis |>
            tibble::as_tibble() |>
            dplyr::left_join(chroms, by = c(seqnames1 = "seqnames2")) |>
            dplyr::rename(cumlength_x = cumlength) |>
            dplyr::left_join(chroms, by = c(seqnames2 = "seqnames2")) |>
            dplyr::rename(cumlength_y = cumlength) |>
            dplyr::mutate(
                x = floor(end1 - (end1 - start1) / 2) + cumlength_x,
                y = floor(end2 - (end2 - start2) / 2) + cumlength_y
            )
        mat <- dplyr::mutate(
            mat, 
            score = ifelse(score > M, M, ifelse(score < m, m, score))
        )
        mat <- rbind(
            mat, 
            mat |> 
                dplyr::mutate(x2 = y, y = x, x = x2) |> 
                dplyr::select(-x2)
        )

        ## -- Plot matrix
        p <- ggMatrix(mat, cols = cmap, limits = limits, grid = show_grid) +
            plotFun +
            ggplot2::labs(
                x = "Genome coordinates",
                y = "Genome coordinates"
            ) + 
            ggplot2::scale_x_continuous(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6), position = 'top') +
            ggplot2::scale_y_reverse(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6))
        if (chrom_lines) {
            p <- p +
                ggplot2::geom_hline(
                    yintercept = chroms$cumlength[-1], 
                    colour = "black", alpha = 0.75, linewidth = 0.15
                ) +
                ggplot2::geom_vline(
                    xintercept = chroms$cumlength[-1], 
                    colour = "black", alpha = 0.75, linewidth = 0.15
                )
        }

        ## -- Add tracks 
        if (length(p_tracks$layers)) {
            p <- p + 
            ggplot2::theme(
                plot.margin = grid::unit(c(5.5, 5.5, 5.5 + 10*ntracks, 5.5), "points")
            ) + patchwork::inset_element(
                p_tracks, 
                left = 0, 
                bottom = -0.1*ntracks, 
                right = 1, 
                top = 0, 
                align_to = 'panel', 
                clip = FALSE
            ) 
        }
        
    }

    p
})

#' @rdname plotMatrix
#' @export

setMethod("plotMatrix", "matrix", function(
    x, 
    scale = 'log10', 
    limits = NULL, 
    dpi = 500, 
    rasterize = TRUE, 
    cmap = NULL  
) {    

    mat <- x
    has_negative_scores <- any(mat < 0, na.rm = TRUE)

    ## -- Scale score
    if (scale == 'log10') {
        mat <- log10(mat)
    } 
    else if (scale == 'log2') {
        mat <- log2(mat)
    }
    else if (scale == 'exp0.2') {
        mat <- mat^0.2
    }
    else if (scale == 'linear') {
        mat <- mat
    }

    ## -- Set matrix limits
    if (!is.null(limits)) {
        m <- limits[1]
        M <- limits[2]
    }
    else {
        mat0 <- mat
        sdiag(mat0, 0) <- NA
        .scores <- mat0
        .scores <- .scores[!is.na(.scores)]
        .scores <- .scores[!is.infinite(.scores)]
        M <- max(.scores)
        m <- min(.scores)
        limits <- c(m, M)
    }

    ## -- Choose color map 
    if (is.null(cmap)) {
        if (has_negative_scores) {
            cmap <- bgrColors()
        }
        else {
            cmap <- coolerColors()
        }
    }
    
    # -- Define plotting approach
    if (rasterize) {
        plotFun <- ggrastr::geom_tile_rast(
            raster.dpi = dpi
        )
    }
    else {
        plotFun <- ggplot2::geom_tile()
    }

    ## -- Convert gis to table and extract x/y
    colnames(mat) <- paste0("bin", seq_len(ncol(mat)))
    mat <- tibble::as_tibble(mat) |>
        dplyr::mutate(x = seq_len(ncol(mat))) |>
        tidyr::pivot_longer(-x, names_to = 'y', values_to = 'score') |> 
        dplyr::mutate(y = gsub('bin', '', y) |> as.numeric()) |>
        tidyr::drop_na(score)

    ## -- Clamp scores to limits
    mat <- dplyr::mutate(mat, score = scales::oob_squish(score, c(m, M)))

    ## -- Plot matrix
    p <- ggMatrix(mat, cols = cmap, limits = limits) +
        plotFun +
        ggplot2::labs(
            x = "Bin number",
            y = "Bin number"
        ) + 
        ggplot2::scale_x_continuous(expand = c(0, 0), position = 'top') +
        ggplot2::scale_y_reverse(expand = c(0, 0))

    p
})

#' @rdname plotMatrix
#' @export

setMethod("plotMatrix", "AggrHiCExperiment", function(
    x, 
    use.scores = 'balanced', 
    scale = 'log10', 
    maxDistance = NULL, 
    loops = NULL, 
    borders = NULL, 
    limits = NULL, 
    dpi = 500, 
    rasterize = TRUE, 
    chrom_lines = TRUE, 
    show_grid = FALSE, 
    cmap = NULL, 
    caption = TRUE  
 ) {
    pairs <- topologicalFeatures(x, 'targets')
    is1D <- all(S4Vectors::first(pairs) == S4Vectors::second(pairs))
    p <- plotMatrix(
        interactions(x), 
        use.scores = use.scores, 
        scale = scale, 
        maxDistance = maxDistance, 
        loops = loops, 
        borders = borders, 
        limits = limits, 
        dpi = dpi, 
        rasterize = rasterize, 
        symmetrical = ifelse(is1D, TRUE, FALSE), 
        chrom_lines = chrom_lines, 
        show_grid = show_grid, 
        cmap = cmap  
    )
    if (caption) {
        p <- p + ggplot2::labs(
            caption = paste(
                sep = '\n',
                paste0('file: ', fileName(x)), 
                paste0('resolution: ', resolution(x)), 
                paste0('focus: ', focus(x)),
                paste0('scores: ', use.scores),
                paste0('scale: ', scale)
            )
        )
    }
    return(p)
})

#' @rdname plotMatrix
#' @export

setMethod("montage", "AggrHiCExperiment", function(
    x, 
    use.scores = 'balanced', 
    scale = 'log10', 
    limits = NULL, 
    dpi = 500, 
    rasterize = TRUE, 
    cmap = NULL
 ) {
    pairs <- topologicalFeatures(x, 'targets')
    is1D <- ifelse(is(pairs, 'GRanges'), TRUE, FALSE)
    sli <- slices(x, use.scores)
    snips <- lapply(seq_len(dim(sli)[3]), function(K){
        mat <- sli[, , K]
        if (is1D) mat[lower.tri(mat)] <- t(mat)[lower.tri(mat)]
        plotMatrix(
            mat, 
            scale = scale, 
            limits = limits, 
            dpi = dpi, 
            rasterize = rasterize, 
            cmap = cmap
        ) + 
            ggplot2::theme_void() + 
            ggplot2::theme(legend.position = "none") + 
            ggplot2::theme(plot.margin = margin(.002, .002, .002, .002, "npc"))
    })
    p <- patchwork::wrap_plots(snips)
})

ggMatrix <- function(mat, ticks = TRUE, grid = FALSE, cols = coolerColors(), limits) {
    p <- ggplot2::ggplot(mat, ggplot2::aes(x, y, fill = score))
    p <- p + ggplot2::scale_fill_gradientn(
        colors = cols,
        na.value = "#FFFFFF",
        limits = limits
    ) +
        ggplot2::guides(fill = ggplot2::guide_colorbar(barheight = ggplot2::unit(.1, "npc"), barwidth = 0.5, frame.colour = "black")) + 
        ggplot2::coord_fixed() +
        ggthemeHiContacts(ticks, grid)
    p
}

ggHorizontalMatrix <- function(df, cols = coolerColors(), limits) {
    p <- ggplot2::ggplot(df, ggplot2::aes(x, diag, fill = score))
    r <- 1/sqrt(2)/sqrt(2)
    p <- p + ggplot2::scale_fill_gradientn(
        colors = cols,
        na.value = "#FFFFFF",
        limits = limits
    ) +
        ggplot2::scale_x_continuous(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6)) +
        ggplot2::scale_y_continuous(expand = c(0, 0), labels = scales::unit_format(unit = "M", scale = 1e-6)) +
        ggplot2::guides(fill = ggplot2::guide_colorbar(barheight = ggplot2::unit(.1, "npc"), barwidth = 0.5, frame.colour = "black")) + 
        ggplot2::coord_fixed(ratio = r) +
        ggthemeHiContacts(ticks = TRUE, grid = FALSE)
    p
}

################################################################################
#                                                                              #
#                                 P(s)                                         #
#                                                                              #
################################################################################

#' Plotting a P(s) distance law
#' 
#' @name plotPs
#' @aliases plotPs
#' @aliases plotPsSlope
#' 
#' @param x the output data.frame of `distanceLaw` function
#' @param mapping aes to pass on to ggplot2
#' @param xlim xlim
#' @param ylim ylim
#' @return ggplot
#'
#' @importFrom scales trans_breaks
#' @importFrom scales trans_format
#' @examples 
#' ## Single P(s)
#' 
#' contacts_yeast <- contacts_yeast()
#' ps <- distanceLaw(contacts_yeast)
#' plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p))
#' 
#' ## Comparing several P(s)
#' 
#' contacts_yeast <- contacts_yeast()
#' contacts_yeast_eco1 <- contacts_yeast_eco1()
#' ps_wt <- distanceLaw(contacts_yeast)
#' ps_wt$sample <- 'WT'
#' ps_eco1 <- distanceLaw(contacts_yeast_eco1)
#' ps_eco1$sample <- 'eco1'
#' ps <- rbind(ps_wt, ps_eco1)
#' plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p, group = sample, color = sample))
#' plotPsSlope(ps, ggplot2::aes(x = binned_distance, y = slope, group = sample))
NULL

#' @rdname plotPs
#' @export

plotPs <- function(x, mapping, xlim = c(5000, 4.99e5), ylim = c(1e-8, 1e-4)) {
    p <- ggplot2::ggplot(x, mapping = mapping) + 
        ggplot2::geom_line() + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(panel.border = ggplot2::element_rect(fill = NA)) + 
        ggplot2::theme(panel.grid.minor = ggplot2::element_blank()) +
        ggplot2::annotation_logticks() + 
        ggplot2::labs(x = "Genomic distance", y = "Contact frequency")
    if (!is.null(ylim)) {
        p <- p + ggplot2::scale_y_log10(
            limits = ylim, 
            expand = c(0, 0), 
            breaks = scales::trans_breaks("log10", function(x) 10^x),
            labels = scales::trans_format("log10", scales::math_format(10^.x))
        )
    }
    else {
        p <- p + ggplot2::scale_y_log10(
            expand = c(0, 0), 
            breaks = scales::trans_breaks("log10", function(x) 10^x),
            labels = scales::trans_format("log10", scales::math_format(10^.x))
        )
    }
    if (!is.null(xlim)) {
        p <- p + ggplot2::scale_x_log10(
            limits = xlim, 
            expand = c(0, 0), 
            breaks = c(1, 10, 100, 1000, 10000, 1e+05, 1e+06, 1e+07, 1e+08, 1e+09, 1e+10),
            labels = c('1', '10', '100', '1kb', '10kb', '100kb', '1Mb', '10Mb', '100Mb', '1Gb', '10Gb')
        )
    }
    else {
        p <- p + ggplot2::scale_x_log10(
            expand = c(0, 0), 
            breaks = c(1, 10, 100, 1000, 10000, 1e+05, 1e+06, 1e+07, 1e+08, 1e+09, 1e+10),
            labels = c('1', '10', '100', '1kb', '10kb', '100kb', '1Mb', '10Mb', '100Mb', '1Gb', '10Gb')
        )
    }
    p
}

#' @rdname plotPs
#' @export

plotPsSlope <- function(x, mapping, xlim = c(5000, 4.99e5), ylim = c(-3, 0)) {
    gg <- ggplot2::ggplot(x, mapping = mapping) + 
        ggplot2::geom_line() + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(panel.border = ggplot2::element_rect(fill = NA)) + 
        ggplot2::theme(panel.grid.minor = ggplot2::element_blank()) +
        ggplot2::scale_y_continuous(
            limits = ylim, 
            expand = c(0, 0)
        ) +
        ggplot2::scale_x_log10(
            limits = xlim, 
            expand = c(0, 0), 
            breaks = c(1, 10, 100, 1000, 10000, 1e+05, 1e+06, 1e+07, 1e+08, 1e+09, 1e+10),
            labels = c('1', '10', '100', '1kb', '10kb', '100kb', '1Mb', '10Mb', '100Mb', '1Gb', '10Gb')
        ) + 
        ggplot2::annotation_logticks(sides = "b") + 
        ggplot2::labs(x = "Genomic distance", y = "Slope of P(s)")
    gg
}

################################################################################
#                                                                              #
#                                 virtual 4C                                   #
#                                                                              #
################################################################################

#' Plotting virtual 4C profiles
#' 
#' @name plot4C
#' @aliases plot4C
#' 
#' @param x GRanges, generally the output of `virtual4C()`
#' @param mapping aes to pass on to ggplot2 (default: 
#' `ggplot2::aes(x = center, y = score, col = seqnames)`)
#' @return ggplot
#' 
#' @import tibble
#' @importFrom scales unit_format
#' @examples 
#' contacts_yeast <- contacts_yeast()
#' v4C <- virtual4C(contacts_yeast, GenomicRanges::GRanges('II:490001-510000'))
#' plot4C(v4C)
NULL

#' @rdname plot4C
#' @export

plot4C <- function(x, mapping = ggplot2::aes(x = center, y = score, col = seqnames)) {
    x <- tibble::as_tibble(x)
    p <- ggplot2::ggplot(x, mapping = mapping) + 
        # ggplot2::geom_line() + 
        ggplot2::geom_area(position = "identity", alpha = 0.2) + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(panel.border = ggplot2::element_rect(fill = NA)) + 
        ggplot2::theme(panel.grid.minor = ggplot2::element_blank()) +
        ggplot2::labs(x = "Position", y = "Contacts with viewpoint") + 
        ggplot2::scale_x_continuous(
            expand = c(0, 0), 
            labels = scales::unit_format(unit = "M", scale = 1e-6)
        )
    p
}

################################################################################
#                                                                              #
#                                 Scalograms                                   #
#                                                                              #
################################################################################

#' Plotting scalograms
#' 
#' @name plotScalogram
#' @aliases plotScalogram
#' 
#' @param x GRanges, the output of `scalogram()`
#' @param ylim Range of distances to use for y-axis in scalograms
#' @return ggplot
#' 
#' @import tibble
#' @examples 
#' contacts_yeast <- HiCExperiment::contacts_yeast()
#' pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
#'   'yeast_wt', format = 'pairs.gz'
#' )
#' scalo <- scalogram(contacts_yeast['II'])
#' plotScalogram(scalo)
NULL

#' @rdname plotScalogram
#' @export

plotScalogram <- function(x, ylim = c(5e2, 1e5)) {
    x_ <- dplyr::mutate(x, dist_quantile = scales::oob_squish(dist_quantile, c(ylim[[1]], ylim[[2]]))) |>
        dplyr::mutate(y0 = ylim[[1]])
    p <- ggplot2::ggplot() + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(panel.border = ggplot2::element_rect(fill = NA)) + 
        ggplot2::theme(panel.grid.minor = ggplot2::element_blank()) +
        ggplot2::scale_x_continuous(expand = c(0, 0)) + 
        ggplot2::scale_y_log10(
            expand = c(0, 0), 
            limits = c(ylim[[1]], ylim[[2]])
        ) + 
        ggplot2::annotation_logticks(sides = 'l') + 
        ggplot2::labs(x = 'Position along chr.', y = 'Distance of interactions')
    if (length(unique(x_$prob)) == 3) {
        x_$prob <- as.numeric(as.character(x_$prob))
        low <- sort(unique(x_$prob))[1]
        mid <- sort(unique(x_$prob))[2]
        high <- sort(unique(x_$prob))[3]
        x_$prob <- dplyr::case_when(
            x_$prob == low ~ 'low', 
            x_$prob == mid ~ 'mid', 
            x_$prob == high ~ 'high'
        )
        diff <- {high - low} * 100
        p <- p + 
            ggplot2::geom_ribbon(
                data = tidyr::pivot_wider(x_[x_$prob != 'mid',], names_from = prob, values_from = dist_quantile), 
                mapping = ggplot2::aes(x = binned_pos, ymin = low, ymax = high, fill = y0), 
                col = 'black', linewidth = 0.2, linetype = 'dashed'
            ) + 
            ggplot2::geom_line(
                data = x_[x_$prob == 'mid',], 
                ggplot2::aes(x = binned_pos, y = dist_quantile, col = prob)
            ) + 
            ggplot2::scale_fill_gradient(low = '#b5b5b54d', high = '#b5b5b54d') + 
            ggplot2::scale_color_manual(values = '#000000') + 
            ggplot2::labs(fill = paste0(diff, '% (', low, '% - ', high, '%)\nof interactions\nwithin this range')) + 
            ggplot2::labs(col = paste0(mid*100, '% of interaction distance')) + 
            ggplot2::theme(legend.text = ggplot2::element_blank(), legend.position = 'bottom')
    } 
    else {
        p <- p + 
            ggplot2::geom_ribbon(data = x_, ggplot2::aes(
                x = binned_pos, ymax = dist_quantile,
                group = prob, fill = prob
            )) + 
            ggplot2::aes(ymin = y0) + 
            ggplot2::scale_fill_brewer(palette = 'Spectral') 
    }
    if (length(unique(x_$chr)) > 1) {
        p <- p + ggplot2::facet_wrap(~chr)
    }
    p
}

################################################################################
#                                                                              #
#                                 Saddle plots                                 #
#                                                                              #
################################################################################

#' Plotting saddle plots
#' 
#' @name plotSaddle
#' @aliases plotSaddle
#' 
#' @param x a HiCExperiment object with a stored `eigens` metadata
#' @param nbins Number of bins to use to discretize the eigenvectors
#' @param limits limits for color map being used
#' @param plotBins Whether to plot the distribution of bins on top of the plot
#' @param BPPARAM a BiocParallel registered method 
#' @return ggplot
NULL

#' @rdname plotSaddle
#' @importFrom GenomicRanges resize
#' @export

plotSaddle <- function(x, nbins = 50, limits = c(-1, 1), plotBins = FALSE, BPPARAM = BiocParallel::bpparam()) {
    if (is.null(metadata(x)$eigens)) stop(
        "No eigen vector found in metadata. Run getCompartments(x) first."
    )
    eigens <- metadata(x)$eigens

    ## -- Filter and bin regions by their eigenvector score
    eigens$eigen_bin <- cut(
        eigens$eigen, breaks = stats::quantile(
            eigens$eigen[eigens$eigen != 0], 
            probs = seq(0.025, 0.975, length.out = nbins+1)
        ), 
        include.lowest = TRUE
    ) |> as.numeric()

    ## -- Compute over-expected score in each pair of eigen bins
    if (interactive()) 
        BiocParallel::bpprogressbar(BPPARAM) <- TRUE
    df <- BiocParallel::bplapply(
        BPPARAM = BPPARAM, 
        as.character(unique(seqnames(eigens))), 
        function(chr) {.saddle(x[chr], eigens)}
    ) |> dplyr::bind_rows()
    dat <- df |> 
        dplyr::group_by(eigen_bin1, eigen_bin2) |> 
        dplyr::summarize(score = mean(detrended), .groups = 'drop') |> 
        dplyr::mutate(
            x = eigen_bin1 / nbins,
            y = eigen_bin2 / nbins, 
            squished_score = scales::oob_squish(score, limits)
        )

    ## -- Make saddle plot 
    p1 <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) + 
        ggrastr::geom_tile_rast(ggplot2::aes(fill = squished_score)) + 
        ggplot2::scale_y_reverse(labels = scales::percent, expand = c(0, 0)) + 
        ggplot2::scale_x_continuous(
            labels = scales::percent, expand = c(0, 0), 
            position = 'top'
        ) + 
        ggplot2::coord_fixed() + 
        ggplot2::scale_fill_gradientn(
            colors = bgrColors(),
            na.value = "#FFFFFF",
            limits = limits
        ) + 
        ggplot2::labs(
            x = '', 
            y = 'Genomic bins ranked by eigenvector value', 
            fill = "Interaction frequency\n(log2 over expected)"
        ) + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(legend.position = 'bottom') +
        ggplot2::theme(panel.border = ggplot2::element_rect(fill = NA)) + 
        ggplot2::theme(axis.title.x = ggplot2::element_blank()) 
    if (!plotBins) return(p1)
    
    ## -- Make side barplot 
    dat2 <- tibble::as_tibble(eigens) |> 
        dplyr::group_by(eigen_bin) |> 
        dplyr::summarize(mean_eigen = mean(eigen)) |> 
        dplyr::mutate(x = eigen_bin / nbins)
    p2 <- ggplot2::ggplot(dat2, ggplot2::aes(x = x, y = mean_eigen)) + 
        ggplot2::geom_col() + 
        ggplot2::scale_x_continuous(
            labels = NULL, expand = c(0, 0)
        ) + 
        ggplot2::labs(
            x = '', 
            y = 'Eigen', 
            fill = "Interaction frequency\n(log2 over expected)"
        ) + 
        ggplot2::theme_minimal() + 
        ggplot2::theme(legend.position = 'bottom') +
        ggplot2::theme(panel.border = ggplot2::element_blank()) + 
        ggplot2::theme(panel.grid = ggplot2::element_blank()) + 
        ggplot2::theme(axis.text.x = ggplot2::element_blank()) +
        ggplot2::theme(axis.title.x = ggplot2::element_blank()) +
        ggplot2::theme(axis.line.x = ggplot2::element_blank()) 
    
    p <- patchwork::wrap_plots(p2, p1, ncol = 1, heights = c(1, 3))

}

.saddle <- function(x_chr, eigens) {
    dx <- detrend(x_chr)
    ints <- interactions(dx)
    an <- anchors(ints)
    ints2 <- ints[IRanges::`%over%`(an[[1]], eigens) & IRanges::`%over%`(an[[2]], eigens)]
    an2 <- anchors(ints2)
    df <- as(ints2, 'data.frame')
    df$eigen_bin1 <- eigens[S4Vectors::subjectHits(
        GenomicRanges::findOverlaps(
            GenomicRanges::resize(an2[[1]], fix = 'center', width = 1), eigens
        )
    )]$eigen_bin
    df$eigen_bin2 <- eigens[S4Vectors::subjectHits(
        GenomicRanges::findOverlaps(
            GenomicRanges::resize(an2[[2]], fix = 'center', width = 1), eigens
        )
    )]$eigen_bin
    drop_na(df, eigen_bin1, eigen_bin2, detrended) |> 
        dplyr::select(eigen_bin1, eigen_bin2, detrended)
}

################################################################################
#                                                                              #
#                                 ggplot2 extra                                #
#                                                                              #
################################################################################

ggthemeHiContacts <- function(ticks = TRUE, grid = FALSE) {
    t <- ggplot2::theme_bw() +
        ggplot2::theme(text = ggplot2::element_text(size = 8))
    if (ticks) t <- t + ggplot2::theme(axis.ticks = ggplot2::element_line(colour = "black", linewidth = 0.2))
    if (grid) {
        t <- t + 
            ggplot2::theme(
                panel.grid.minor = ggplot2::element_line(linewidth = 0.025, colour = "#00000052"), 
                panel.grid.major = ggplot2::element_line(linewidth = 0.025, colour = "#00000052")
            )
            t
    } else {
        t <- t + 
            ggplot2::theme(
                panel.grid.minor = ggplot2::element_blank(), 
                panel.grid.major = ggplot2::element_blank()
            )
            t
    }
}

#' Matrix palettes
#' 
#' @name palettes
#' 
#' @return A vector of colours carefully
#' picked for Hi-C contact heatmaps
#' 
#' @examples
#' bwrColors()
#' bbrColors()
#' bgrColors()
#' afmhotrColors()
#' coolerColors()
#' rainbowColors()
NULL

#' @rdname palettes
#' @export

bwrColors <- function() {
    c("#1659b1", "#4778c2", "#ffffff", "#b13636", "#6C150E")
}

#' @rdname palettes
#' @export

bbrColors <- function() {
    c("#1659b1", "#4778c2", "#a9c3e7", "#ffffff", 
        "#e2adad", "#b13636", "#6C150E"
    )
}

#' @rdname palettes
#' @export

bgrColors <- function() {
    rev(c(
        "#BD202D",
        "#DE614D",
        "#F19374",
        "#F6BA9E",
        "#EAD5CB",
        "#CEDBEB",
        "#AFC5E6",
        "#8FA7D6",
        "#677CBD",
        "#495BA9"
    ))
}

#' @rdname palettes
#' @export

afmhotrColors <- function() {
    c(
        "#ffffff", 
        "#f8f5c3", 
        "#f4ee8d", 
        "#f6be35", 
        "#ee7d32",
        "#c44228", 
        "#821d19", 
        "#381211", 
        "#050606"
    )
}

#' @rdname palettes
#' @export

coolerColors <- function() {
    rev(c(
        "#1A0A10",
        "#7A1128",
        "#B01F29",
        "#D42027",
        "#ED3024",
        "#F15C34",
        "#F78E40",
        "#FAAA4B",
        "#FFCA67",
        "#FDE188",
        "#FFF2A9",
        "#FCF9CE",
        "#FFFEF9"
    ))
}

#' @rdname palettes
#' @export

rainbowColors <- function() {
    rev(c(
        "#FDEE00",
        "#FEC00F",
        "#F37421",
        "#ED1E24",
        "#F26768",
        "#F9B7B8",
        "#FFFCFC",
        "#C6E8ED",
        "#8BD1EC",
        "#56ABDF",
        "#3864AF",
        "#3A57A7",
        "#8D509F"
    ))
}

test_that("compartments works", {
    library(BSgenome.Scerevisiae.UCSC.sacCer3)
    genome <- BSgenome.Scerevisiae.UCSC.sacCer3
    GenomeInfoDb::seqlevelsStyle(genome) <- "NCBI"
    full_contacts_yeast <- HiCExperiment::contacts_yeast(full = TRUE)
    comps <- getCompartments(full_contacts_yeast, genome = genome)
    comps_VI <- getCompartments(full_contacts_yeast, chromosomes = "VI")
    expect_no_error(getCompartments(full_contacts_yeast, 
        genome = genome, 
        chromosomes = "VI"
    ))
    expect_s4_class(
        comps_VI, 
        'HiCExperiment'
    )
    expect_s4_class(
        HiCExperiment::topologicalFeatures(comps_VI, 'compartments'), 
        'GRanges'
    )
    expect_s4_class(
        S4Vectors::metadata(comps_VI)$eigens, 
        'GRanges'
    )
    expect_s3_class(
        plotSaddle(comps_VI), 'gg'
    )
})

test_that("insulation works", {
    hic <- HiCExperiment::contacts_yeast() |> 
        HiCExperiment::refocus('II:1-300000') |> 
        HiCExperiment::zoom(resolution = 1000)
    hic2 <- getDiamondInsulation(hic, window_size = 8000)
    
    expect_no_error(getDiamondInsulation(hic, window_size = 8000))
    expect_no_error(getBorders(hic2))
    expect_s4_class(
        hic2, 
        'HiCExperiment'
    )
    expect_s4_class(
        HiCExperiment::topologicalFeatures(hic2, 'borders'), 
        'GRanges'
    )
    expect_s4_class(
        S4Vectors::metadata(hic2)$insulation, 
        'GRanges'
    )
})

test_that("scalogram works", {
    contacts_yeast <- HiCExperiment::contacts_yeast()
    pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
        'yeast_wt', format = 'pairs.gz'
    )
    scalo1 <- scalogram(contacts_yeast['II'])
    scalo2 <- scalogram(contacts_yeast['II'], probs = c(0, 0.3, 1))
    expect_no_error(scalogram(contacts_yeast['II']))
    expect_no_error(scalogram(contacts_yeast['II'], probs = c(0, 0.3, 1)))
    expect_s3_class(plotScalogram(scalo1), 'gg')
    expect_s3_class(plotScalogram(scalo2), 'gg')

})

test_that("plotMatrix works", {
    contacts_yeast <- HiCExperiment::contacts_yeast()
    full_contacts_yeast <- HiCExperiment::contacts_yeast()
    cm <- Matrix::as.matrix(gi2cm(interactions(full_contacts_yeast), 'balanced'))
    y <- base::as.matrix(cm)
    expect_s3_class(plotMatrix(y), 'gg')
    expect_s3_class(plotMatrix(contacts_yeast), 'gg')
    expect_s3_class(plotMatrix(contacts_yeast, scale = 'linear'), 'gg')
    expect_s3_class(plotMatrix(contacts_yeast, scale = 'log2'), 'gg')
    expect_s3_class(
        plotMatrix(
            contacts_yeast, 
            scale = 'log10', 
            limits = c(-1, 1), 
            cmap = bwrColors()
        ),
        'gg'
    )
    expect_s3_class(
        plotMatrix(
            full_contacts_yeast,
            scale = 'exp0.2', 
            limits = c(-1, 1), 
            cmap = bbrColors()
        ),
        'gg'
    )
    expect_s3_class(
        plotMatrix(
            full_contacts_yeast['II:1-800000'],
            borders = topologicalFeatures(full_contacts_yeast, 'centromeres'),
            scale = 'exp0.2', 
            limits = c(-1, 1), 
            cmap = bbrColors(), 
            show_grid = TRUE
        ),
        'gg'
    )
})

#' @title Aligning tracks with HiCExperiment objects
#' @name tracks
#' @rdname tracks
#' @description
#' 
#' Aligning tracks with HiCExperiment objects
#'
#' @param x A `HiCExperiment` object over a full genome
#' @param use.pairs logical. Whether to use pairsFile to compute Hi-C coverage
#' @param bin.size if `use.pairs == TRUE`, to which resolution 
#' @return A `HiCExperiment` object with 2 added columns in `regions(x)`
#' 
#' @examples 
#' mcool_file <- HiContactsData::HiContactsData('yeast_wt', format = 'mcool')
#' hic <- import(mcool_file, format = 'mcool', resolution = 1000)
#' coverage(hic)
NULL

#' @param rle a bigwig track imported as 'Rle'
#' @param track.name Name to give to the track in regions metadata
#' @noRd

alignTrack <- function(x, rle, track.name = 'track') {
    re <- regions(interactions(x))
    mc <- S4Vectors::mcols(re)
    mc[[track.name]] <- rle[re]
    mc[[paste0(track.name, '.mean')]] <- BiocGenerics::mean(
        mc[[track.name]], na.rm = TRUE
    )
    S4Vectors::mcols(regions(interactions(x))) <- mc
    return(x)
}

#' @rdname tracks
#' @importFrom GenomicRanges tileGenome
#' @importFrom GenomicRanges countOverlaps
#' @importFrom IRanges coverage
#' @export

setMethod("coverage", signature(x = "HiCExperiment"), function(x, use.pairs = FALSE, bin.size = resolution(x)) {
    pairsFile <- HiCExperiment::pairsFile(x)
    
    ## -- Define tiles on which to count coverage
    if (bin.size > resolution(x)) {
        tiles <- GenomicRanges::tileGenome(
            seqinfo(x), tilewidth = bin.size, cut.last.tile.in.chrom = TRUE
        )
    } 
    else {
        tiles <- bins(x) |> as("GRanges")
    }

    if (use.pairs & !is.null(pairsFile)) { ## Pairs provided: easy countOverlaps
        message("Importing pairs file ", pairsFile, " in memory. This may take a while...")
        pairs <- BiocIO::import(pairsFile, format = 'pairs') 
        an <- anchors(pairs)
        reads <- c(an[[1]], an[[2]])
        tiles$cnt <- GenomicRanges::countOverlaps(tiles, reads)
        tiles$CPM <- tiles$cnt / sum(tiles$cnt) * 1e6
        IRanges::coverage(tiles, weight = 'CPM')
    } 
    else { ## Pairs not provided: recover each ends from binned ints, findOverlaps
        ints <- interactions(x)
        an <- anchors(ints)
        an[[1]]$cnt <- ints$count
        an[[2]]$cnt <- ints$count
        reads <- c(an[[1]], an[[2]])
        df <- dplyr::left_join(
            as.data.frame(reads), 
            as.data.frame(tiles),
            by = dplyr::join_by(seqnames, start, end)
        ) |> dplyr::group_by(seqnames, start, end) |> 
            dplyr::summarize(cnt = sum(cnt), .groups = 'drop') 
        tiles <- GenomicRanges::makeGRangesFromDataFrame(df, keep.extra.columns = TRUE)
        tiles$CPM <- tiles$cnt / sum(tiles$cnt) * 1e6
        IRanges::coverage(tiles, weight = 'CPM')
    }

})

`sdiag<-` <- function(A, k = 0, value) {
    p <- ncol(A)
    n <- nrow(A)
    if (k>p-1||-k > n-1) return()
    if (k >= 0) {
        i <- seq_len(n)
        j <- (k+1):p
    } 
    else {
        i <- (-k+1):n
        j <- seq_len(p)
    }
    if (length(i)>length(j)) i <- i[seq_along(j)] else j <- j[seq_along(i)]
    ii <- i + (j-1) * n 
    A[ii] <- value
    A
} 

#' @rdname arithmetics
#' @export

boost <- function(x, use.scores = 'balanced', alpha = 1, full.replace = FALSE) {

    if (!requireNamespace("Rfast", quietly = TRUE)) {
        message("Install Rfast package to perform Boost-HiC.")
        message("install.packages('Rfast')")
    }
    gis <- HiCExperiment::interactions(x)
    gis$score <- HiCExperiment::scores(x, use.scores)
    cm <- HiCExperiment::gi2cm(gis)
    an <- InteractionSet::anchors(cm)
    mat <- HiCExperiment::cm2matrix(cm)
    
    d0 <- (1/(mat^(1/alpha)))
    distances <- Rfast::floyd(d0)
    F <- (1/(distances^(1/alpha)))

    if (!full.replace) {
        mat_boosted <- mat
        mat_boosted[is.na(mat) | mat == 0] <- F[is.na(mat) | mat == 0]
    }
    else {
        mat_boosted <- F
    }

    cm_boosted <- InteractionSet::ContactMatrix(
        mat_boosted, 
        anchor1 = an[[1]], 
        anchor2 = an[[2]], 
        regions = InteractionSet::regions(cm)
    )
    is_boosted <- InteractionSet::deflate(cm_boosted)
    gis_boosted <- InteractionSet::interactions(is_boosted)
    gis_boosted$bin_id1 <- HiCExperiment::anchors(gis_boosted)[[1]]$bin_id
    gis_boosted$bin_id2 <- HiCExperiment::anchors(gis_boosted)[[2]]$bin_id
    HiCExperiment::interactions(x) <- gis_boosted
    m <- dplyr::left_join(
        as.data.frame(GenomicRanges::mcols(gis_boosted)), 
        as.data.frame(GenomicRanges::mcols(gis)), 
        by = c('bin_id1', 'bin_id2')
    ) |> dplyr::select(-bin_id1, -bin_id2, -score)
    m$boosted <- SummarizedExperiment::assay(is_boosted, 1)[, 1]
    l <- as.list(m) |> S4Vectors::SimpleList()
    x@scores <- l
    metadata(x)[['alpha']] <- alpha

    return(x)
}

#' cisTransRatio
#' 
#' Quickly computes a cis-trans ratio of interactions. 
#'
#' @param x A `HiCExperiment` object over the full genome
#' @return a tibble, listing for each chr. the % of cis/trans interactions
#'
#' @importFrom tibble as_tibble
#' @importFrom tidyr pivot_wider
#' @importFrom dplyr rename
#' @importFrom dplyr relocate
#' @importFrom dplyr group_by
#' @importFrom dplyr summarize
#' @importFrom dplyr mutate
#' @export
#' @examples 
#' library(HiContacts)
#' full_contacts_yeast <- contacts_yeast(full = TRUE)
#' cisTransRatio(full_contacts_yeast)

cisTransRatio <- function(x) {
    if (!is.null(focus(x))) {
        stop('Please provide a contact matrix over the entire genome. Aborting now.')
    }
    gis <- InteractionSet::interactions(x)
    gis$score <- HiCExperiment::scores(x, 'count')
    cnts <- gis |> 
        tibble::as_tibble() |> 
        dplyr::relocate(c(seqnames1, seqnames2))
    cnts_dup <- cnts |> 
        dplyr::rename(seqnames1 = seqnames2, seqnames2 = seqnames1) |> 
        dplyr::relocate(c(seqnames1, seqnames2))
    cnts <- rbind(cnts, cnts_dup)
    res <- cnts |> 
        dplyr::group_by(seqnames1, seqnames2) |> 
        dplyr::summarize(n = sum(score, na.rm = TRUE)) |> 
        dplyr::mutate(type = ifelse(seqnames1 == seqnames2, 'cis', 'trans')) |> 
        dplyr::group_by(seqnames1, type) |> 
        dplyr::rename(chr = seqnames1) |>
        dplyr::summarize(n = sum(n, na.rm = TRUE)) |> 
        tidyr::pivot_wider(names_from = type, values_from = n) |> 
        dplyr::mutate(
            n_total = sum(cis + trans, na.rm = TRUE), 
            cis_pct = cis/n_total, 
            trans_pct = trans/n_total
        )
    return(res)
}

#' Computing virtual 4C profiles
#' 
#' From a (m)cool pre-imported in memory, computes a 4C profile 
#' using a user-specified `viewpoint`. 
#'
#' @param x a `HiCExperiment` object
#' @param viewpoint viewpoint, defined as a GRanges
#' @param use.scores use.scores
#' @return A tibble with the contact frequency of the viewpoint, per bin 
#'   along the imported genomic range.
#' 
#' @import tibble
#' @importFrom scales unit_format
#' @importFrom S4Vectors queryHits
#' @importFrom IRanges IRanges
#' @importFrom GenomicRanges GRanges
#' @importFrom GenomicRanges start
#' @importFrom GenomicRanges end
#' @importFrom GenomicRanges findOverlaps
#' @export
#' @examples 
#' library(HiContacts)
#' contacts_yeast <- contacts_yeast()
#' v4C <- virtual4C(contacts_yeast, GenomicRanges::GRanges('II:490001-510000'))
#' v4C

virtual4C <- function(x, viewpoint, use.scores = 'balanced') {
    gis <- InteractionSet::interactions(x)
    gis$score <- HiCExperiment::scores(x, use.scores)
    cm <- Matrix::as.matrix(gi2cm(gis))
    cm <- base::as.matrix(cm)
    regions <- InteractionSet::regions(gis)
    regions_in_viewpoint <- seq_along(regions) %in% S4Vectors::queryHits(
        GenomicRanges::findOverlaps(regions, viewpoint)
    )
    if (sum(regions_in_viewpoint) > 1) {
        score <- rowSums(cm[, regions_in_viewpoint], na.rm = TRUE)
    } 
    else {
        score <- cm[, regions_in_viewpoint]
    }
    gr <- GenomicRanges::GRanges(
        seqnames = as.vector(GenomicRanges::seqnames(regions)),
        IRanges::IRanges(
            GenomicRanges::start(regions), 
            GenomicRanges::end(regions)
        ), 
        score = score,
        center = GenomicRanges::start(regions) + 
            (GenomicRanges::end(regions) - GenomicRanges::start(regions))/2, 
        in_viewpoint = regions_in_viewpoint
    )
    if (length(as.character(viewpoint)) == 1) {
        gr$viewpoint <- as.character(viewpoint)
    }
    return(gr)
}

#' HiContacts arithmetics functionalities
#' @name arithmetics
#' @rdname arithmetics
#' @export

setMethod("aggregate", signature(x = "HiCExperiment"), function(
    x, 
    targets, 
    flankingBins = 51, 
    maxDistance = NULL,
    BPPARAM = BiocParallel::bpparam()
) {
    HiCExperiment::AggrHiCExperiment(
        file = fileName(x), 
        resolution = resolution(x), 
        targets = targets,  
        flankingBins = flankingBins, 
        metadata = S4Vectors::metadata(x), 
        topologicalFeatures = topologicalFeatures(x), 
        pairsFile = pairsFile(x), 
        maxDistance = maxDistance, 
        BPPARAM = BPPARAM,
        bed = metadata(x)$regions
    )
})

#' @importFrom utils txtProgressBar
#' @importFrom utils setTxtProgressBar
#' @importFrom stats median
#' @rdname arithmetics
#' @aliases normalize,HiCExperiment-method
#' @export

setMethod("normalize", signature(object = "HiCExperiment"), function(
    object, use.scores = 'count', niters = 200, min.nnz = 10, mad.max = 3
) {

    x <- object
    gis <- HiCExperiment::interactions(x)
    gis$score <- HiCExperiment::scores(x, use.scores)
    cm <- HiCExperiment::gi2cm(gis)
    an <- InteractionSet::anchors(cm)
    re <- InteractionSet::regions(cm)
    mat0 <- HiCExperiment::cm2matrix(cm, sparse = TRUE)
    mat0 <- Matrix::triu(mat0)
    mat <- mat0

    # - Filter bins 
    bins <- Matrix::colSums(mat)
    bins[bins == 0] <- 1
    lbins <- log10(bins)
    med <- stats::median(lbins)
    sig <- 1.4826 * stats::mad(lbins)
    s_min <- med - mad.max * sig
    filtered_bins <- {lbins > s_min} & 
        {Matrix::colSums(mat > 0, na.rm = TRUE) > min.nnz}
    invalid_bins <- re[which(!filtered_bins)]$bin_id

    # - Iterate over max_iter
    mat_iced <- mat[filtered_bins, filtered_bins]
    nonzeros <- data.frame(col = mat_iced@i+1, row = mat_iced@j+1) |> as.matrix()
    if (interactive()) 
        pb <- utils::txtProgressBar(min = 0, max = niters, style = 3, width = 50, char = "-") 
    for (i in seq_len(niters)) {
        if (interactive()) utils::setTxtProgressBar(pb, i)
        bin_sums <- Matrix::colSums(mat_iced, na.rm = TRUE)
        bin_sums <- bin_sums / stats::median(bin_sums, na.rm = TRUE)
        mat_iced@x <- mat_iced@x / {bin_sums[nonzeros[, 'row']] * bin_sums[nonzeros[, 'col']]}
    }
    cat('\n')

    # - Rescale ice-d scores
    bin_sums <- Matrix::colSums(mat_iced, na.rm = TRUE)
    mat_iced@x <- mat_iced@x / stats::median(bin_sums, na.rm = TRUE)

    # - Make upper-tri
    mat_iced <- Matrix::triu(mat_iced)

    # - Recover scores 
    cm_iced <- InteractionSet::ContactMatrix(
        mat_iced, 
        anchor1 = an[[1]][filtered_bins],
        anchor2 = an[[2]][filtered_bins],
        regions = re[filtered_bins]
    )
    is_iced <- InteractionSet::deflate(cm_iced)
    gis_iced <- InteractionSet::interactions(is_iced)
    gis_iced$bin_id1 <- HiCExperiment::anchors(gis_iced)[[1]]$bin_id
    gis_iced$bin_id2 <- HiCExperiment::anchors(gis_iced)[[2]]$bin_id
    gis_iced$ICE <- SummarizedExperiment::assay(is_iced, 1)[, 1]
    m <- dplyr::left_join(
        as.data.frame(GenomicRanges::mcols(gis)), 
        as.data.frame(GenomicRanges::mcols(gis_iced)), 
        by = c('bin_id1', 'bin_id2')
    )
    m$ICE[{m$bin_id1 %in% invalid_bins} | {m$bin_id2 %in% invalid_bins}] <- NA
    m <- dplyr::select(m, -bin_id1, -bin_id2, -score)

    # - Save new scores
    x@scores <- as.list(m) |> S4Vectors::SimpleList()

    return(x)
})

#' @title Finding loops in contact map
#' @name getLoops
#' @rdname getLoops
#' @description
#' 
#' Find loops using chromosight.  
#' 
#' This function is actually provided by the `HiCool` package rather than 
#' the `HiContacts` package. `HiCool` provides a self-managed
#' `conda` environment, and this limits 
#' 
#' @param ... Parameters passed to `HiCool::getLoops()`.
#' @export

getLoops <- function(...){
    stop("`getLoops` is provided by the `HiCool` package. Please install HiCool and use `HiCool::getLoops()`.")
}

utils::globalVariables(c(
    ".",
    ".x",
    "as",
    "average_interaction_per_diag",
    "binned_distance",
    "binned_pos",
    "binwidth",
    "centros_yeast",
    "chr",
    "cis",
    "column_center",
    "coords_chr",
    "coords_end",
    "coords_start",
    "cumlength",
    "cumlength_x",
    "cumlength_y",
    "data", 
    "dist",
    "dist_column",
    "dist_quantile",
    "dist_row",
    "distance",
    "dists",
    "end1",
    "end2",
    "group",
    "is",
    "ints",
    "loess",
    "n",
    "n_total",
    "ninter",
    "norm_p",
    "norm_p_unity",
    "p",
    "path",
    "pos", 
    "pos_",
    "predict",
    "prob",
    "row_center",
    "score",
    "seqnames1",
    "seqnames2",
    "serpentine_ncores",
    "serpentine_niter",
    "setNames",
    "slope",
    "sm1",
    "sm2",
    "start1",
    "start2",
    "term",
    "threshold",
    "trans",
    "trend",
    "type",
    "x",
    "x2",
    "y",
    "y0", 
    "center1",
    "center2",
    "cm_mat",
    "corr",
    "detrended",
    "eigen_bin",
    "eigen_bin1",
    "eigen_bin2",
    "id",
    "l",
    "mean_eigen",
    "quantile",
    "region",
    "squished_score", 
    "bin_id1", 
    "bin_id2", 
    "center", 
    "margin", 
    "mcols", 
    "score.x", 
    "score.by", 
    "reduce", 
    "value", 
    "cnt", 
    "fc"
))

#' @rdname plotMatrix
#' @param x Can be a `HiCExperiment`, a GInteractions or an array/matrix object. 
#' @param ... Extra arguments passed to the corresponding method.

setGeneric("plotMatrix", function(x, ...) {standardGeneric("plotMatrix")})

#' @rdname plotMatrix

setGeneric("montage", function(x, ...) {standardGeneric("montage")})

#' @rdname Ps

setGeneric("distanceLaw", function(x, coords, ...) {standardGeneric("distanceLaw")})

#' @title Checks functions
#' 
#' @name checks
#' @rdname checks
#' 
#' @description 
#' 
#' Useful functions to validate the nature/structure of (m)cool files or 
#' `HiCExperiment` objects.
#'  All these check functions should return a logical.
#' 
#' @param x A `HiCExperiment` object
#' @param ... `HiCExperiment` objects
#' @return Logical
#' @keywords internal
NULL

#' @rdname checks

.is_symmetrical <- function(x) {
    if (is.null(focus(x))) {
        return(TRUE)
    }
    if (grepl('\\|', focus(x))) {
        return(FALSE)
    }
    else {
        return(TRUE)
    }
}

#' @rdname checks

.is_comparable <- function(...) {
    .are_HiCExperiment(...)
    err <- c()
    if (!.is_same_seqinfo(...)) {
        err <- c(err, "seqinfos")
    }
    if (!.is_same_resolution(...)) {
        err <- c(err, "resolutions")
    }
    if (!.is_same_bins(...)) {
        err <- c(err, "bins")
    }
    if (!.is_same_regions(...)) {
        err <- c(err, "regions")
    }
    if (length(err) > 0) {
        mess <- paste0("Provided `HiCExperiment` have different ", paste(err, collapse = ' & '), '.')
        stop(mess)
    }
    TRUE
}

#' @rdname checks

.are_HiCExperiment <- function(...) {
    args <- list(...)
    if (!all(unlist(lapply(args, is, 'HiCExperiment')))) {
        stop("Provided arguments are not all `HiCExperiment` objects. 
        Please only use `HiCExperiment` objects with this function.")
    }
    TRUE
}

#' @rdname checks

.is_same_seqinfo <- function(...) {
    contacts_list <- list(...)
    all(unlist(lapply(contacts_list, function(x) {
        identical(seqinfo(contacts_list[[1]]), seqinfo(x))
    })))
}

#' @rdname checks

.is_same_resolution <- function(...) {
    contacts_list <- list(...)
    all(unlist(lapply(contacts_list, function(x) {
        identical(resolution(contacts_list[[1]]), resolution(x))
    })))
}

#' @rdname checks

.is_same_bins <- function(...) {
    contacts_list <- list(...)
    all(unlist(lapply(contacts_list, function(x) {
        b1 <- bins(contacts_list[[1]])
        b1$weight <- NULL
        b2 <- bins(x)
        b2$weight <- NULL
        identical(b1, b2)
    })))
}

#' @rdname checks

.is_same_regions <- function(...) {
    contacts_list <- list(...)
    all(unlist(lapply(contacts_list, function(x) {
        re1 <- regions(contacts_list[[1]])
        re1$weight <- NULL
        re2 <- regions(x)
        re2$weight <- NULL
        identical(re1, re2)
    })))
}

test_that("Contacts works", {
    contacts_yeast <- HiCExperiment::contacts_yeast()

    expect_warning(
        Contacts(fileName(contacts_yeast), focus = 'II', resolution = 16000)
    )
    
})

test_that("v4C works", {
    contacts_yeast <- HiCExperiment::contacts_yeast() |> 
        HiCExperiment::refocus("II")
    expect_s4_class(
        virtual4C(contacts_yeast, GRanges('II:495001-505000')),
        'GRanges'
    )
    expect_s3_class({
        v4C <- virtual4C(contacts_yeast, GRanges('II:495001-505000'))
        plot4C(v4C, ggplot2::aes(x = center, y = score))
    }, 'gg')
})

test_that("cistrans works", {
    contacts_yeast <- HiCExperiment::contacts_yeast()
    expect_s3_class({
        cisTransRatio(contacts_yeast)
    }, 'tbl')
})

test_that("Ps works", {
    contacts_yeast <- HiCExperiment::contacts_yeast()
    x1 <- distanceLaw(contacts_yeast)
    x2 <- localDistanceLaw(contacts_yeast, GRanges("II:1-100000"))
    pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
        'yeast_wt', format = 'pairs.gz'
    )
    y1 <- distanceLaw(contacts_yeast, by_chr = TRUE)
    y2 <- localDistanceLaw(contacts_yeast, GRanges("II:1-100000"))
    expect_s3_class(x1, 'tbl')
    expect_s3_class(y1, 'tbl')
    expect_s3_class(x2, 'tbl')
    expect_s3_class(y2, 'tbl')
    expect_s3_class(plotPs(x1, ggplot2::aes(x = binned_distance, y = norm_p)), 'gg')
    expect_s3_class(plotPsSlope(x1, ggplot2::aes(x = binned_distance, y = norm_p)), 'gg')
})

test_that("checks and utils work", {

    expect_true(is.matrix({
        m <- matrix(data = 0, nrow = 100, ncol = 100)
        sdiag(m, k = 0) <- 3
        m
    }))

    contacts_yeast <- HiCExperiment::contacts_yeast() |> 
        HiCExperiment::refocus("II:1-30000")
    contacts_yeast_eco1 <- HiCExperiment::contacts_yeast_eco1() |> 
        HiCExperiment::refocus("II:1-30000")

    expect_true(.is_symmetrical(contacts_yeast_eco1))
    expect_true(.is_comparable(contacts_yeast, contacts_yeast_eco1))
    expect_true(.are_HiCExperiment(contacts_yeast, contacts_yeast_eco1))
    expect_true(.is_same_seqinfo(contacts_yeast, contacts_yeast_eco1))
    expect_true(.is_same_resolution(contacts_yeast, contacts_yeast_eco1))
    expect_true(.is_same_bins(contacts_yeast, contacts_yeast_eco1))
    expect_true(.is_same_regions(contacts_yeast, contacts_yeast_eco1))

})

#' @title Contact map insulation
#' @name getDiamondInsulation
#' @rdname getDiamondInsulation
#' @description
#' 
#' Computes diamond insulation score along the entire genome 
#'
#' @param x A `HiCExperiment` object over a full genome
#' @param window_size Which window size to use to compute diamond insulation 
#' score (default: 10 * resolution)
#' @param weak_threshold Less stringent cutoff to call borders in the 
#' diamond insulation score
#' @param strong_threshold More stringent cutoff to call borders in the 
#' diamond insulation score
#' @param BPPARAM BiocParallel parallelization settings 
#' @return a `HiCExperiment` object with additional `insulation` metadata, 
#' containing the diamond insulation score computed 
#'
#' @importFrom Matrix sparseMatrix
#' @export
#' @examples 
#' library(HiContacts)
#' hic <- contacts_yeast() |> 
#'   refocus('II:1-300000') |> 
#'   zoom(1000)
#' diams <- getDiamondInsulation(hic)
#' getDiamondInsulation(diams)

getDiamondInsulation <- function(
    x, 
    window_size = NULL, 
    BPPARAM = BiocParallel::bpparam()
) {

    message( "Going through preflight checklist..." )
    # - Check resolutions 
    si <- seqinfo(x)
    resolution <- resolution(x)
    if(is.null(window_size)) window_size <- 8 * resolution
    ress <- resolutions(x)
    if (!resolution %in% ress) stop("Chosen resolution not available")
    if (!window_size %in% ress) stop("Chosen window_size not available")
    fact <- window_size / resolution
    cm <- gi2cm(interactions(x), 'balanced')
    cm_mat <- as.matrix(cm)

    # - Scan each bin individually
    message( "Scan each window and compute diamond insulation score..." )
    idx_ids <- seq(fact+1, nrow(cm)-fact-1)
    scores <- BiocParallel::bplapply(BPPARAM = BPPARAM, 
        idx_ids, 
        function(idx) {
            # Extract matrix from chr-level, centered on idx
            mat <- cm_mat[seq(idx-fact, idx+fact), seq(idx-fact, idx+fact)]
            # Extract top-right quadrant
            mat_trq <- mat[seq(1, fact) + 1 - 1, seq(fact, fact + fact - 1) + 1 + 1]
            sum(mat_trq,na.rm = TRUE)
        }
    ) |> unlist()
    fbins <- regions(cm)[idx_ids]
    fbins$score <- scores
    fbins$insulation <- log2( scores / mean(scores) )
    fbins <- fbins[!is.na(fbins$insulation)]

    # - Annotate boundaries
    message( "Annotating diamond score prominence for each window..." )
    mins <- which(diff(sign(diff(fbins$insulation)))==+2)+1
    maxs <- which(diff(sign(diff(fbins$insulation)))==-2)+1
    if (length(maxs) < length(mins)) maxs[[length(maxs)+1]] <- length(fbins)
    if (length(maxs) > length(mins)) maxs <- utils::tail(maxs, -1)
    fbins$min <- seq_along(fbins) %in% mins
    fbins$prominence <- NA
    fbins$prominence[fbins$min] <- -{fbins$insulation[fbins$min] - fbins$insulation[seq_along(fbins) %in% maxs]}
    
    ## -- Return initial HiCExperiment object with insulation in metadata
    metadata(x)$insulation <- fbins
    return(x)

}

#' @rdname getDiamondInsulation
#' @export

getBorders <- function(x, weak_threshold = 0.2, strong_threshold = 0.5) {

    fbins <- metadata(x)$insulation
    fbins$border <- fbins$prominence >= weak_threshold & fbins$min
    fbins$type <- ifelse(fbins$prominence < weak_threshold, NA, ifelse(fbins$prominence < strong_threshold, 'weak', 'strong'))
    gr <- fbins[which(fbins$border)]
    gr$score <- gr$prominence
    names(gr) <- gr$type
    S4Vectors::mcols(gr) <- data.frame(score = gr$score)
    topologicalFeatures(x, 'borders') <- gr
    return(x)

}

.getDiamondScore <- function(cm, bin, fact) {
    ids <- seq(bin$bin_id - fact, bin$bin_id + fact)
    idx_ids <- seq(fact+1, nrow(cm_mat)-fact-1)
    balanced <- l[ids+1, ids+1]
    s <- sum(
        balanced[seq(1, fact) + 1 - 1, seq(fact, fact + fact - 1) + 1 + 1],
        na.rm = TRUE
    )
    return(s)
}

#' HiContacts arithmetics functionalities
#' 
#' @name arithmetics
#' @aliases detrend
#' @aliases autocorrelate
#' @aliases divide
#' @aliases merge,HiCExperiment,HiCExperiment-method
#' @aliases despeckle
#' @aliases aggregate,HiCExperiment-method
#' @aliases boost
#' @aliases subsample
#' @aliases coarsen
#' @aliases normalize,HiCExperiment-method
#' 
#' @description 
#' Different arithmetic operations can be performed on Hi-C contact matrices:  
#'  - `normalize` a contact matrix using iterative correction;
#'  - `detrend` a contact matrix, i.e. remove the distance-dependent 
#' contact trend;
#'  - `autocorrelate` a contact matrix: this is is typically done to highlight 
#' large-scale compartments;
#'  - `divide` one contact matrix by another; 
#'  - `merge` multiple contact matrices;
#'  - `despeckle` (i.e. smooth out) a contact matrix out;
#'  - `aggregate` (average) a contact matrices over a set of genomic loci of 
#' interest;
#'  - `boost` Hi-C signal by enhancing long-range interactions while preserving short-
#' range interactions (this is adapted from Boost-HiC);
#'  - `subsample` interactions using a proportion or a fixed number of final 
#' interactions.
#'  - `coarsen` a contact matrix to a larger (coarser) resolution
#' 
#' @param x,y,object a `HiCExperiment` object
#' @param use.scores Which scores to use to perform operations
#' @param ... `HiCExperiment` objects. For `aggregate`, `targets` (a set of 
#' GRanges or GInteractions).
#' @param detrend Detrend matrix before performing autocorrelation
#' @param ignore_ndiags ignore N diagonals when calculating correlations
#' @param by a `HiCExperiment` object
#' @param focal.size Size of the smoothing rectangle
#' @param alpha Power law scaling factor. As indicated in Boost-HiC documentation, 
#' the alpha parameter influences the weighting of contacts: if alpha < 1 
#' long-range interactions are prioritized; if alpha >> 1 short-range 
#' interactions have more weight when computing the distance matrix.
#' @param full.replace Whether to replace the entire set of contacts, 
#' rather than only filling the missing interactions (count=0) (Default: FALSE)
#' @param prop Float between 0 and 1, or integer corresponding to the # of 
#' @param niters Number of iterations for ICE matrix balancing
#' @param min.nnz Filter bins with less than `min.nnz` non-zero elements when 
#' performing ICE matrix balancing
#' @param mad.max Filter out bins whose log coverage is less than `mad.max` 
#' median absolute deviations below the median bin log coverage.
#' @param targets Set of chromosome coordinates for which 
#'   interaction counts are extracted from the Hi-C contact file, provided
#'   as a GRanges object (for diagnoal-centered loci) or as a GInteractions
#'   object (for off-diagonal coordinates).
#' @param pseudocount Add a pseudocount when dividing matrices (Default: 0)
#' @param bin.size Bin size to coarsen a HiCExperiment at
#' @param flankingBins Number of bins on each flank of the bins containing 
#'   input targets.
#' @param maxDistance Maximum distance to use when compiling distance decay
#' @param FUN merging function
#' @param BPPARAM BiocParallel parameters
#' 
#' @return a `HiCExperiment` object with extra scores
#' 
#' @import InteractionSet
#' @import stringr
#' @import tidyr
#' @importFrom BiocGenerics normalize
#' @importFrom scales rescale
#' @importFrom scales oob_censor
#' @importFrom stats quantile
#' @importFrom tibble as_tibble
#' @importFrom tibble tibble
#' @importFrom dplyr group_by
#' @importFrom dplyr summarize
#' @importFrom dplyr pull
#' @importFrom dplyr mutate
#' @importFrom dplyr left_join
#' @importFrom dplyr full_join
#' @importFrom dplyr rename
#' @importFrom dplyr n
#' @importFrom GenomicRanges seqnames
#' @importFrom GenomicRanges findOverlaps
#' @importFrom GenomicRanges start
#' @importFrom GenomicRanges GRanges
#' @importFrom S4Vectors metadata
#' @importFrom S4Vectors SimpleList
#' @importFrom S4Vectors mcols
#' @importFrom SummarizedExperiment assay
#' @importFrom BiocParallel bpparam
#' 
#' @examples 
#' #### -----
#' #### Normalize a contact matrix
#' #### -----
#' 
#' library(HiContacts)
#' contacts_yeast <- contacts_yeast()
#' normalize(contacts_yeast)
#'
#' #### -----
#' #### Detrending a contact matrix
#' #### -----
#' 
#' detrend(contacts_yeast)
#'
#' #### -----
#' #### Auto-correlate a contact matrix
#' #### -----
#' 
#' autocorrelate(contacts_yeast)
#' 
#' #### -----
#' #### Divide 2 contact matrices
#' #### -----
#' 
#' contacts_yeast <- refocus(contacts_yeast, 'II')
#' contacts_yeast_eco1 <- contacts_yeast_eco1() |> refocus('II')
#' divide(contacts_yeast_eco1, by = contacts_yeast)
#' 
#' #### -----
#' #### Merge 2 contact matrices
#' #### -----
#' 
#' merge(contacts_yeast_eco1, contacts_yeast)
#' 
#' #### -----
#' #### Despeckle (smoothen) a contact map
#' #### -----
#' 
#' despeckle(contacts_yeast)
#' 
#' #### -----
#' #### Aggregate a contact matrix over centromeres, at different scales
#' #### -----
#' 
#' contacts <- contacts_yeast() |> zoom(resolution = 1000)
#' centros <- topologicalFeatures(contacts, 'centromeres')
#' aggregate(contacts, targets = centros, flankingBins = 51)
#' 
#' #### -----
#' #### Enhance long-range interaction signal
#' #### -----
#' 
#' contacts <- contacts_yeast() |> zoom(resolution = 1000) |> refocus('II')
#' boost(contacts, alpha = 1)
#' 
#' #### -----
#' #### Subsample & "coarsen" contact matrix 
#' #### -----
#' 
#' subcontacts <- subsample(contacts, prop = 100000)
#' coarsened_subcontacts <- coarsen(subcontacts, bin.size = 4000)
NULL

#' @rdname arithmetics
#' @export

detrend <- function(x, use.scores = 'balanced') {
    gis <- interactions(x)
    gis$score <- scores(x, use.scores)
    gis$diag <- gis$bin_id2 - gis$bin_id1
    if (is.null(metadata(x)[['detrending_model']])) {
        expected <- tibble::as_tibble(gis) |> 
            dplyr::group_by(diag) |> 
            dplyr::mutate(score = scales::oob_censor(
                score, range = stats::quantile(score, c(0.05, 0.95), na.rm = TRUE)
            )) |> 
            dplyr::summarize(average_interaction_per_diag = mean(score, na.rm = TRUE)) |> 
            dplyr::mutate(average_interaction_per_diag = average_interaction_per_diag / 2)
    }
    else {
        expected <- metadata(x)[['detrending_model']]
        expected$average_interaction_per_diag <- expected$score
        expected$distance <- NULL
        expected$score <- NULL
    }
    gis$expected <- tibble::as_tibble(gis) |> 
        dplyr::left_join(expected, by = 'diag') |> 
        dplyr::pull(average_interaction_per_diag)
    gis$score_over_expected <- log2( {gis$score/sum(gis$score, na.rm = TRUE)} / {gis$expected/sum(gis$expected, na.rm = TRUE)} )
    # gis$score_over_expected <- log2( {scale(gis$score)} / {scale(gis$expected)} )[,1]
    scores(x, "expected") <- gis$expected
    scores(x, "detrended") <- gis$score_over_expected
    return(x)
}

#' @rdname arithmetics
#' @export

autocorrelate <- function(x, use.scores = 'balanced', detrend = TRUE, ignore_ndiags = 3) {
    if (!requireNamespace("WGCNA", quietly = TRUE)) {
        message("Install WGCNA package to perform autocorrelation.")
        message("install.packages('WGCNA')")
    }
    if (detrend) {
        if (!{'detrend' %in% names(scores(x))}) {
            x <- detrend(x, use.scores = use.scores)
        }
        use.scores <- 'detrended'
    }
    gis <- interactions(x)
    gis$score <- scores(x, use.scores)
    reg <- regions(gis)
    mat <- cm2matrix(gi2cm(gis))
    for (K in seq(-ignore_ndiags, ignore_ndiags, by = 1)) {
        sdiag(mat, K) <- NA
    }
    co <- WGCNA::cor(mat, use = 'pairwise.complete.obs', method = "pearson")
    colnames(co) <- names(reg)
    rownames(co) <- names(reg)
    mat2 <- tibble::as_tibble(co, rownames = 'region', .name_repair = "minimal") |> 
        tidyr::pivot_longer(-region, names_to = "y", values_to = "corr") |>
        dplyr::rename("x" = "region")
    mat2$bin_id1 <- dplyr::left_join(
        mat2, 
        data.frame(region = names(reg), id = reg$bin_id), 
        by = c(x = 'region')
    ) |> pull(id)
    mat2$bin_id2 <- dplyr::left_join(
        mat2, 
        data.frame(region = names(reg), id = reg$bin_id), 
        by = c(y = 'region')
    ) |> pull(id)
    scores(x, 'autocorrelated') <- left_join(
        as_tibble(gis), mat2, by = c("bin_id1", "bin_id2")
    ) |> pull(corr)
    return(x)
}

#' @rdname arithmetics
#' @export

divide <- function(x, by, use.scores = 'balanced', pseudocount = 0) {

    x_gis <- interactions(x)
    x_gis$score <- S4Vectors::mcols(x_gis)[[use.scores]]
    by_gis <- interactions(by)
    by_gis$score <- S4Vectors::mcols(by_gis)[[use.scores]]

    ## -- If regions are different, manually merge them 
    re <- unique(
        c(InteractionSet::regions(x_gis), InteractionSet::regions(by_gis))
    )
    InteractionSet::replaceRegions(x_gis) <- re
    InteractionSet::replaceRegions(by_gis) <- re
    
    ## -- Check that all objects are comparable (bins, resolution, seqinfo)
    .is_same_seqinfo(x, by)
    .is_same_resolution(x, by)
    .is_same_bins(x, by)

    ## -- Join tables, divide, and convert back to GInteractions
    gis <- dplyr::full_join(
        as.data.frame(x_gis) |> dplyr::select(!contains(c('.1', 'chr', 'width', 'strand', 'center'))), 
        as.data.frame(by_gis) |> dplyr::select(!contains(c('.1', 'chr', 'width', 'strand', 'center'))),
        by = c(
            'seqnames1', 'start1', 'end1', 'bin_id1', 
            'seqnames2', 'start2', 'end2', 'bin_id2'
        ), 
        suffix = c(".x", ".by")
    ) |> 
        dplyr::mutate(
            score.x = score.x + pseudocount, 
            score.by = score.by + pseudocount, 
            fc = score.x / score.by, 
            l2fc = log2(fc)
        ) |> 
        HiCExperiment::df2gi()
    S4Vectors::mcols(gis)[[paste0(use.scores, '.fc')]] <- gis$fc
    S4Vectors::mcols(gis)[[paste0(use.scores, '.l2fc')]] <- gis$l2fc
    S4Vectors::mcols(gis)[, c('score.x', 'score.by', 'fc', 'l2fc')] <- NULL
    InteractionSet::replaceRegions(gis) <- re

    ## -- Export results as HiCExperiment
    m <- S4Vectors::mcols(gis) |> 
        as.data.frame() |> 
        dplyr::select(!starts_with('weight'))
    S4Vectors::mcols(gis) <- m
    m <- dplyr::select(m, -c('bin_id1', 'bin_id2'))
    scores <- as.list(m) |> S4Vectors::SimpleList()

    ## -- Create HiCExperiment
    res <- methods::new("HiCExperiment", 
        focus = HiCExperiment::focus(x), 
        metadata = list(
            hce_list = list(x = x, by = by), 
            operation = 'divide'
        ),
        resolutions = resolution(x), 
        resolution = resolution(x), 
        interactions = gis, 
        scores = scores, 
        topologicalFeatures = S4Vectors::SimpleList(), 
        pairsFile = NULL, 
        fileName = ""
    )
    return(res)

}

#' @rdname arithmetics
#' @export

setMethod("merge", signature(x = "HiCExperiment", y = "HiCExperiment"), definition = function(
    x, y, ..., use.scores = 'balanced', FUN = mean) 
{

    contacts_list <- list(x, y, ...)
    
    ## -- Check that at least 2 `HiCExperiment` objects are passed to `merge()`
    .are_HiCExperiment(x, y, ...)
    if (length(contacts_list) < 2) {
        stop("Please provide at least 2 `HiCExperiment` objects.")
    } 

    ## -- Check that all objects are comparable (bins, regions, resolution, seqinfo)
    .is_same_seqinfo(x, y, ...)
    .is_same_resolution(x, y, ...)
    .is_same_bins(x, y, ...)

    # Unify all the interactions
    score_names <- names(scores(contacts_list[[1]]))
    ints <- do.call(
        c, lapply(contacts_list, FUN = interactions) 
    ) |> sort()
    re <- do.call(
        c, lapply(contacts_list, FUN = regions) 
    ) |> sort() |> unique()
    ints_df <- as.data.frame(ints)
    merged_ints <- dplyr::select(ints_df, !any_of(score_names)) |> 
        dplyr::distinct() |> 
        HiCExperiment:::asGInteractions()
    replaceRegions(merged_ints) <- re

    # Group by bin_id1/bin_id2 and merge scores
    if (identical(FUN, mean)) {
        .FUN <- function(x) mean(x, na.rm = TRUE)
    }
    else if (identical(FUN, sum)){
        .FUN <- function(x) sum(x, na.rm = TRUE)
    }
    scores <- dplyr::select(ints_df, dplyr::any_of(c('bin_id1', 'bin_id2', score_names))) |> 
        dplyr::group_by(bin_id1, bin_id2) |>
        dplyr::summarise(dplyr::across(dplyr::all_of(score_names), .FUN), .groups = "drop") |> 
        dplyr::select(-bin_id1, -bin_id2) |> 
        as("SimpleList")

    ## -- Create a new HiCExperiment object
    files <- unlist(lapply(contacts_list, fileName))
    res <- methods::new("HiCExperiment", 
        focus = focus(contacts_list[[1]]), 
        metadata = list(
            hce_list = contacts_list, 
            operation = .FUN
        ),
        resolutions = resolutions(contacts_list[[1]]), 
        resolution = resolution(contacts_list[[1]]), 
        interactions = merged_ints, 
        scores = scores, 
        topologicalFeatures = S4Vectors::SimpleList(), 
        pairsFile = NULL, 
        fileName = fileName(contacts_list[[1]])
    )
    return(res)

})

#' @rdname arithmetics
#' @export

despeckle <- function(x, use.scores = 'balanced', focal.size = 1) {
    if (!requireNamespace("terra", quietly = TRUE)) {
        message("Install terra package to despeckle contact matrix.")
        message("install.packages('terra')")
    }

    gis <- HiCExperiment::interactions(x)
    gis$score <- HiCExperiment::scores(x, use.scores)
    cm <- HiCExperiment::gi2cm(gis)
    an <- InteractionSet::anchors(cm)
    mat <- HiCExperiment::cm2matrix(cm)
    r <- terra::rast(mat)
    gauss <- matrix(data = 1/{(focal.size*2+1)*(focal.size*2+1)-1}, ncol = focal.size*2+1, nrow = focal.size*2+1)
    gauss[focal.size+1, focal.size+1] <- 0
    mat_despeckled <- terra::focal(
        x = r, 
        w = gauss, 
        fun = sum,
        na.rm = TRUE
    )#/{(focal.size*2+1)*(focal.size*2+1)-1}
    mat_despeckled <- terra::as.array(mat_despeckled)[, , 1]
    cm_despeckled <- InteractionSet::ContactMatrix(
        mat_despeckled, 
        anchor1 = an[[1]], 
        anchor2 = an[[2]], 
        regions = InteractionSet::regions(cm)
    )
    is_despeckled <- InteractionSet::deflate(cm_despeckled)
    gis_despeckled <- InteractionSet::interactions(is_despeckled)
    gis_despeckled$bin_id1 <- HiCExperiment::anchors(gis_despeckled)[[1]]$bin_id
    gis_despeckled$bin_id2 <- HiCExperiment::anchors(gis_despeckled)[[2]]$bin_id
    HiCExperiment::interactions(x) <- gis_despeckled
    m <- dplyr::left_join(
        as.data.frame(S4Vectors::mcols(gis_despeckled)), 
        as.data.frame(S4Vectors::mcols(gis)), 
        by = c('bin_id1', 'bin_id2')
    ) |> dplyr::select(-bin_id1, -bin_id2, -score)
    m[[paste0(use.scores, '.despeckled')]] <- SummarizedExperiment::assay(is_despeckled, 1)[, 1]
    l <- as.list(m) |> S4Vectors::SimpleList()
    x@scores <- l
    return(x)
}

test_that("arithmetics works", {
    contacts_yeast <- HiCExperiment::contacts_yeast() |> 
        HiCExperiment::zoom(4000) |>
        HiCExperiment::refocus('II:1-100000')
    contacts_yeast_eco1 <- HiCExperiment::contacts_yeast_eco1() |> 
        HiCExperiment::zoom(4000) |>
        HiCExperiment::refocus('II:1-100000')
    full_contacts_yeast <- HiCExperiment::contacts_yeast() |> 
        HiCExperiment::zoom(4000) |>
        HiCExperiment::refocus('II:1-100000') |> 
        HiCExperiment::zoom(resolution = 1000)
    
    expect_true({
        x <- normalize(contacts_yeast)
        validObject(x)
    })
    expect_type(scores(x, 'ICE'), 'double')
    
    expect_true({
        x <- detrend(contacts_yeast)
        validObject(x)
    })

    expect_true({
        x <- autocorrelate(contacts_yeast)
        validObject(x)
    })

    expect_true({
        div_contacts <- divide(contacts_yeast_eco1, by = contacts_yeast)
        validObject(div_contacts)
    })

    expect_true({
        merged_contacts <- merge(contacts_yeast_eco1, contacts_yeast)
        validObject(merged_contacts)
    })
    
    expect_true({
        aggr_contacts <- aggregate(
            full_contacts_yeast, 
            targets = topologicalFeatures(full_contacts_yeast, 'centromeres')
        )
        validObject(aggr_contacts)
    })

    expect_true({
        x <- boost(contacts_yeast)
        validObject(x)
    })

    expect_true({
        x <- subsample(contacts_yeast, prop = 0.1)
        validObject(x)
    })
    expect_true({
        x <- subsample(contacts_yeast, prop = 1000)
        validObject(x)
    })
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotScalogram}
\alias{plotScalogram}
\title{Plotting scalograms}
\usage{
plotScalogram(x, ylim = c(500, 1e+05))
}
\arguments{
\item{x}{GRanges, the output of \code{scalogram()}}

\item{ylim}{Range of distances to use for y-axis in scalograms}
}
\value{
ggplot
}
\description{
Plotting scalograms
}
\examples{
contacts_yeast <- HiCExperiment::contacts_yeast()
pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
  'yeast_wt', format = 'pairs.gz'
)
scalo <- scalogram(contacts_yeast['II'])
plotScalogram(scalo)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/plotting.R
\name{plotMatrix}
\alias{plotMatrix}
\alias{montage}
\alias{plotMatrix,HiCExperiment-method}
\alias{plotMatrix,GInteractions-method}
\alias{plotMatrix,matrix-method}
\alias{plotMatrix,AggrHiCExperiment-method}
\alias{montage,AggrHiCExperiment-method}
\title{Plotting a contact matrix}
\usage{
plotMatrix(x, ...)

montage(x, ...)

\S4method{plotMatrix}{HiCExperiment}(
  x,
  compare.to = NULL,
  use.scores = "balanced",
  scale = "log10",
  maxDistance = NULL,
  loops = NULL,
  borders = NULL,
  tracks = NULL,
  limits = NULL,
  dpi = 500,
  rasterize = TRUE,
  symmetrical = TRUE,
  chrom_lines = TRUE,
  show_grid = FALSE,
  cmap = NULL,
  caption = TRUE
)

\S4method{plotMatrix}{GInteractions}(
  x,
  use.scores = NULL,
  scale = "log10",
  maxDistance = NULL,
  loops = NULL,
  borders = NULL,
  tracks = NULL,
  limits = NULL,
  dpi = 500,
  rasterize = TRUE,
  symmetrical = TRUE,
  chrom_lines = TRUE,
  show_grid = FALSE,
  cmap = NULL
)

\S4method{plotMatrix}{matrix}(
  x,
  scale = "log10",
  limits = NULL,
  dpi = 500,
  rasterize = TRUE,
  cmap = NULL
)

\S4method{plotMatrix}{AggrHiCExperiment}(
  x,
  use.scores = "balanced",
  scale = "log10",
  maxDistance = NULL,
  loops = NULL,
  borders = NULL,
  limits = NULL,
  dpi = 500,
  rasterize = TRUE,
  chrom_lines = TRUE,
  show_grid = FALSE,
  cmap = NULL,
  caption = TRUE
)

\S4method{montage}{AggrHiCExperiment}(
  x,
  use.scores = "balanced",
  scale = "log10",
  limits = NULL,
  dpi = 500,
  rasterize = TRUE,
  cmap = NULL
)
}
\arguments{
\item{x}{A HiCExperiment object}

\item{...}{Extra arguments passed to the corresponding method.}

\item{compare.to}{Compare to a second HiC matrix in the lower left corner}

\item{use.scores}{Which scores to use in the heatmap}

\item{scale}{Any of 'log10', 'log2', 'linear', 'exp0.2' (Default: 'log10')}

\item{maxDistance}{maximum distance. If provided, the heatmap is plotted
horizontally}

\item{loops}{Loops to plot on top of the heatmap, provided as \code{GInteractions}}

\item{borders}{Borders to plot on top of the heatmap, provided as \code{GRanges}}

\item{tracks}{Named list of bigwig tracks imported as \code{Rle}}

\item{limits}{color map limits}

\item{dpi}{DPI to create the plot (Default: 500)}

\item{rasterize}{Whether the generated heatmap is rasterized or vectorized
(Default: TRUE)}

\item{symmetrical}{Whether to enforce a symetrical heatmap (Default: TRUE)}

\item{chrom_lines}{Whether to display separating lines between chromosomes,
should any be necessary (Default: TRUE)}

\item{show_grid}{Whether to display an underlying grid (Default: FALSE)}

\item{cmap}{Color scale to use. (Default: bgrColors() if limits are c(-1, 1)
and coolerColors() otherwise)}

\item{caption}{Whether to display a caption (Default: TRUE)}
}
\value{
ggplot object
}
\description{
Plotting a contact matrix
}
\examples{
contacts_yeast <- contacts_yeast()
plotMatrix(
    contacts_yeast, 
    use.scores = 'balanced', 
    scale = 'log10', 
    limits = c(-4, -1)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/4C.R
\name{virtual4C}
\alias{virtual4C}
\title{Computing virtual 4C profiles}
\usage{
virtual4C(x, viewpoint, use.scores = "balanced")
}
\arguments{
\item{x}{a \code{HiCExperiment} object}

\item{viewpoint}{viewpoint, defined as a GRanges}

\item{use.scores}{use.scores}
}
\value{
A tibble with the contact frequency of the viewpoint, per bin
along the imported genomic range.
}
\description{
From a (m)cool pre-imported in memory, computes a 4C profile
using a user-specified \code{viewpoint}.
}
\examples{
library(HiContacts)
contacts_yeast <- contacts_yeast()
v4C <- virtual4C(contacts_yeast, GenomicRanges::GRanges('II:490001-510000'))
v4C
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compartments.R
\name{getCompartments}
\alias{getCompartments}
\title{Contact map compartments}
\usage{
getCompartments(
  x,
  resolution = NULL,
  genome = NULL,
  chromosomes = NULL,
  neigens = 3,
  sort_eigens = FALSE,
  BPPARAM = BiocParallel::bpparam()
)
}
\arguments{
\item{x}{A \code{HiCExperiment} object over a full genome}

\item{resolution}{Which resolution to use to compute eigen vectors}

\item{genome}{a BSgenome of DNAStringSet object associated with the Hi-C
contact matrix.}

\item{chromosomes}{character or integer vector indicating which}

\item{neigens}{Numver of eigen vectors to extract}

\item{sort_eigens}{Can be FALSE or one of c('Spearman', 'Pearson')}

\item{BPPARAM}{BiocParallel parallelization settings}
}
\value{
A \code{HiCExperiment} object with additional \code{eigens} metadata containing the
normalized eigenvectors and a new "compartments" topologicalFeatures
storing A and B compartments as a GRanges object.
}
\description{
Computes eigen vectors for each chromosome using cis contacts and extract
chromosome compartments.
}
\examples{
library(HiContacts)
full_contacts_yeast <- contacts_yeast(full = TRUE)
comps <- getCompartments(full_contacts_yeast)
metadata(comps)$eigens
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plot4C}
\alias{plot4C}
\title{Plotting virtual 4C profiles}
\usage{
plot4C(x, mapping = ggplot2::aes(x = center, y = score, col = seqnames))
}
\arguments{
\item{x}{GRanges, generally the output of \code{virtual4C()}}

\item{mapping}{aes to pass on to ggplot2 (default:
\code{ggplot2::aes(x = center, y = score, col = seqnames)})}
}
\value{
ggplot
}
\description{
Plotting virtual 4C profiles
}
\examples{
contacts_yeast <- contacts_yeast()
v4C <- virtual4C(contacts_yeast, GenomicRanges::GRanges('II:490001-510000'))
plot4C(v4C)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loops.R
\name{getLoops}
\alias{getLoops}
\title{Finding loops in contact map}
\usage{
getLoops(...)
}
\arguments{
\item{...}{Parameters passed to \code{HiCool::getLoops()}.}
}
\description{
Find loops using chromosight.

This function is actually provided by the \code{HiCool} package rather than
the \code{HiContacts} package. \code{HiCool} provides a self-managed
\code{conda} environment, and this limits
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R, R/arithmetics.R, R/boost.R,
%   R/coarsen.R, R/normalize.R, R/subsample.R
\name{arithmetics}
\alias{arithmetics}
\alias{detrend}
\alias{autocorrelate}
\alias{divide}
\alias{merge,HiCExperiment,HiCExperiment-method}
\alias{despeckle}
\alias{aggregate,HiCExperiment-method}
\alias{boost}
\alias{subsample}
\alias{coarsen}
\alias{normalize,HiCExperiment-method}
\title{HiContacts arithmetics functionalities}
\usage{
\S4method{aggregate}{HiCExperiment}(
  x,
  targets,
  flankingBins = 51,
  maxDistance = NULL,
  BPPARAM = BiocParallel::bpparam()
)

detrend(x, use.scores = "balanced")

autocorrelate(x, use.scores = "balanced", detrend = TRUE, ignore_ndiags = 3)

divide(x, by, use.scores = "balanced", pseudocount = 0)

\S4method{merge}{HiCExperiment,HiCExperiment}(x, y, ..., use.scores = "balanced", FUN = mean)

despeckle(x, use.scores = "balanced", focal.size = 1)

boost(x, use.scores = "balanced", alpha = 1, full.replace = FALSE)

coarsen(x, bin.size)

\S4method{normalize}{HiCExperiment}(
  object,
  use.scores = "count",
  niters = 200,
  min.nnz = 10,
  mad.max = 3
)

subsample(x, prop)
}
\arguments{
\item{x, y, object}{a \code{HiCExperiment} object}

\item{targets}{Set of chromosome coordinates for which
interaction counts are extracted from the Hi-C contact file, provided
as a GRanges object (for diagnoal-centered loci) or as a GInteractions
object (for off-diagonal coordinates).}

\item{flankingBins}{Number of bins on each flank of the bins containing
input targets.}

\item{maxDistance}{Maximum distance to use when compiling distance decay}

\item{BPPARAM}{BiocParallel parameters}

\item{use.scores}{Which scores to use to perform operations}

\item{detrend}{Detrend matrix before performing autocorrelation}

\item{ignore_ndiags}{ignore N diagonals when calculating correlations}

\item{by}{a \code{HiCExperiment} object}

\item{pseudocount}{Add a pseudocount when dividing matrices (Default: 0)}

\item{...}{\code{HiCExperiment} objects. For \code{aggregate}, \code{targets} (a set of
GRanges or GInteractions).}

\item{FUN}{merging function}

\item{focal.size}{Size of the smoothing rectangle}

\item{alpha}{Power law scaling factor. As indicated in Boost-HiC documentation,
the alpha parameter influences the weighting of contacts: if alpha < 1
long-range interactions are prioritized; if alpha >> 1 short-range
interactions have more weight when computing the distance matrix.}

\item{full.replace}{Whether to replace the entire set of contacts,
rather than only filling the missing interactions (count=0) (Default: FALSE)}

\item{bin.size}{Bin size to coarsen a HiCExperiment at}

\item{niters}{Number of iterations for ICE matrix balancing}

\item{min.nnz}{Filter bins with less than \code{min.nnz} non-zero elements when
performing ICE matrix balancing}

\item{mad.max}{Filter out bins whose log coverage is less than \code{mad.max}
median absolute deviations below the median bin log coverage.}

\item{prop}{Float between 0 and 1, or integer corresponding to the # of}
}
\value{
a \code{HiCExperiment} object with extra scores
}
\description{
Different arithmetic operations can be performed on Hi-C contact matrices:
\itemize{
\item \code{normalize} a contact matrix using iterative correction;
\item \code{detrend} a contact matrix, i.e. remove the distance-dependent
contact trend;
\item \code{autocorrelate} a contact matrix: this is is typically done to highlight
large-scale compartments;
\item \code{divide} one contact matrix by another;
\item \code{merge} multiple contact matrices;
\item \code{despeckle} (i.e. smooth out) a contact matrix out;
\item \code{aggregate} (average) a contact matrices over a set of genomic loci of
interest;
\item \code{boost} Hi-C signal by enhancing long-range interactions while preserving short-
range interactions (this is adapted from Boost-HiC);
\item \code{subsample} interactions using a proportion or a fixed number of final
interactions.
\item \code{coarsen} a contact matrix to a larger (coarser) resolution
}
}
\examples{
#### -----
#### Normalize a contact matrix
#### -----

library(HiContacts)
contacts_yeast <- contacts_yeast()
normalize(contacts_yeast)

#### -----
#### Detrending a contact matrix
#### -----

detrend(contacts_yeast)

#### -----
#### Auto-correlate a contact matrix
#### -----

autocorrelate(contacts_yeast)

#### -----
#### Divide 2 contact matrices
#### -----

contacts_yeast <- refocus(contacts_yeast, 'II')
contacts_yeast_eco1 <- contacts_yeast_eco1() |> refocus('II')
divide(contacts_yeast_eco1, by = contacts_yeast)

#### -----
#### Merge 2 contact matrices
#### -----

merge(contacts_yeast_eco1, contacts_yeast)

#### -----
#### Despeckle (smoothen) a contact map
#### -----

despeckle(contacts_yeast)

#### -----
#### Aggregate a contact matrix over centromeres, at different scales
#### -----

contacts <- contacts_yeast() |> zoom(resolution = 1000)
centros <- topologicalFeatures(contacts, 'centromeres')
aggregate(contacts, targets = centros, flankingBins = 51)

#### -----
#### Enhance long-range interaction signal
#### -----

contacts <- contacts_yeast() |> zoom(resolution = 1000) |> refocus('II')
boost(contacts, alpha = 1)

#### -----
#### Subsample & "coarsen" contact matrix 
#### -----

subcontacts <- subsample(contacts, prop = 100000)
coarsened_subcontacts <- coarsen(subcontacts, bin.size = 4000)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diamond.R
\name{getDiamondInsulation}
\alias{getDiamondInsulation}
\alias{getBorders}
\title{Contact map insulation}
\usage{
getDiamondInsulation(x, window_size = NULL, BPPARAM = BiocParallel::bpparam())

getBorders(x, weak_threshold = 0.2, strong_threshold = 0.5)
}
\arguments{
\item{x}{A \code{HiCExperiment} object over a full genome}

\item{window_size}{Which window size to use to compute diamond insulation
score (default: 10 * resolution)}

\item{BPPARAM}{BiocParallel parallelization settings}

\item{weak_threshold}{Less stringent cutoff to call borders in the
diamond insulation score}

\item{strong_threshold}{More stringent cutoff to call borders in the
diamond insulation score}
}
\value{
a \code{HiCExperiment} object with additional \code{insulation} metadata,
containing the diamond insulation score computed
}
\description{
Computes diamond insulation score along the entire genome
}
\examples{
library(HiContacts)
hic <- contacts_yeast() |> 
  refocus('II:1-300000') |> 
  zoom(1000)
diams <- getDiamondInsulation(hic)
getDiamondInsulation(diams)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cistrans.R
\name{cisTransRatio}
\alias{cisTransRatio}
\title{cisTransRatio}
\usage{
cisTransRatio(x)
}
\arguments{
\item{x}{A \code{HiCExperiment} object over the full genome}
}
\value{
a tibble, listing for each chr. the \% of cis/trans interactions
}
\description{
Quickly computes a cis-trans ratio of interactions.
}
\examples{
library(HiContacts)
full_contacts_yeast <- contacts_yeast(full = TRUE)
cisTransRatio(full_contacts_yeast)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{checks}
\alias{checks}
\alias{.is_symmetrical}
\alias{.is_comparable}
\alias{.are_HiCExperiment}
\alias{.is_same_seqinfo}
\alias{.is_same_resolution}
\alias{.is_same_bins}
\alias{.is_same_regions}
\title{Checks functions}
\usage{
.is_symmetrical(x)

.is_comparable(...)

.are_HiCExperiment(...)

.is_same_seqinfo(...)

.is_same_resolution(...)

.is_same_bins(...)

.is_same_regions(...)
}
\arguments{
\item{x}{A \code{HiCExperiment} object}

\item{...}{\code{HiCExperiment} objects}
}
\value{
Logical
}
\description{
Useful functions to validate the nature/structure of (m)cool files or
\code{HiCExperiment} objects.
All these check functions should return a logical.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tracks.R
\name{tracks}
\alias{tracks}
\alias{coverage,HiCExperiment-method}
\title{Aligning tracks with HiCExperiment objects}
\usage{
\S4method{coverage}{HiCExperiment}(x, use.pairs = FALSE, bin.size = resolution(x))
}
\arguments{
\item{x}{A \code{HiCExperiment} object over a full genome}

\item{use.pairs}{logical. Whether to use pairsFile to compute Hi-C coverage}

\item{bin.size}{if \code{use.pairs == TRUE}, to which resolution}
}
\value{
A \code{HiCExperiment} object with 2 added columns in \code{regions(x)}
}
\description{
Aligning tracks with HiCExperiment objects
}
\examples{
mcool_file <- HiContactsData::HiContactsData('yeast_wt', format = 'mcool')
hic <- import(mcool_file, format = 'mcool', resolution = 1000)
coverage(hic)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{HiContacts-plots}
\alias{HiContacts-plots}
\title{HiContacts plotting functionalities}
\description{
Several plots can be generated in HiContacts:
\itemize{
\item Hi-C contact matrices
\item Distance-dependant interaction frequency decay (a.k.a. "Distance law" or "P(s)")
\item Virtual 4C profiles
\item Scalograms
\item Saddle plots
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scalogram.R
\name{scalogram}
\alias{scalogram}
\title{Compute a scalogram of contacts}
\usage{
scalogram(x, dist_min = 0, nbins = 250, probs = c(0.25, 0.5, 0.75))
}
\arguments{
\item{x}{A \code{HiCExperiment} object}

\item{dist_min}{Minimum distance for interactions to be considered.}

\item{nbins}{Number of bins to divide each chromosome}

\item{probs}{Quantiles of interactions}
}
\value{
a tibble

a tibble
}
\description{
Compute a scalogram of contacts
}
\examples{
contacts_yeast <- HiCExperiment::contacts_yeast()
pairsFile(contacts_yeast) <- HiContactsData::HiContactsData(
  'yeast_wt', format = 'pairs.gz'
)
scalo <- scalogram(contacts_yeast['II'])
scalo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/Ps.R
\name{distanceLaw}
\alias{distanceLaw}
\alias{Ps}
\alias{distanceLaw,GInteractions,missing-method}
\alias{distanceLaw,HiCExperiment,missing-method}
\alias{distanceLaw,PairsFile,missing-method}
\alias{distanceLaw,PairsFile,GRanges-method}
\alias{localDistanceLaw}
\alias{distanceLaw,HiCExperiment,GRanges-method}
\title{Compute the law of distance-dependent contact frequency, a.k.a. P(s)}
\usage{
distanceLaw(x, coords, ...)

\S4method{distanceLaw}{GInteractions,missing}(x, by_chr = FALSE)

\S4method{distanceLaw}{HiCExperiment,missing}(
  x,
  by_chr = FALSE,
  filtered_chr = c("XII", "chrXII", "chr12", "12", "Mito", "MT", "chrM")
)

\S4method{distanceLaw}{PairsFile,missing}(
  x,
  by_chr = FALSE,
  filtered_chr = c("XII", "chrXII", "chr12", "12", "Mito", "MT", "chrM"),
  chunk_size = 1e+05
)

\S4method{distanceLaw}{HiCExperiment,GRanges}(x, coords, chunk_size = 1e+05)

\S4method{distanceLaw}{PairsFile,GRanges}(x, coords, chunk_size = 1e+05)

localDistanceLaw(x, coords = coords)
}
\arguments{
\item{x}{A \code{HiCExperiment} object}

\item{coords}{GRanges to specify which genomic loci to use when computing P(s)}

\item{...}{Arguments passed to corresponding method}

\item{by_chr}{by_chr}

\item{filtered_chr}{filtered_chr}

\item{chunk_size}{For pairs files which do not fit in memory, pick a number
of pairs to parse by chunks (1e7 should be a good compromise)}
}
\value{
a tibble
}
\description{
P(s) will be approximated if no pairs are provided, or the exact P(s)
will be computed if a \code{.pairs} file is added to the \code{HiCExperiment} object
using \code{pairsFile(x) <- "..."}.
}
\examples{
contacts_yeast <- contacts_yeast()
ps <- distanceLaw(contacts_yeast)
ps
local_ps <- localDistanceLaw(
    contacts_yeast,
    GenomicRanges::GRanges(
        c("telomere" = "II:1-20000", "arm" = "II:300001-700000")
    )
)
local_ps
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contacts.R
\name{Contacts}
\alias{Contacts}
\title{Contacts}
\usage{
Contacts(
  file,
  resolution = NULL,
  focus = NULL,
  metadata = list(),
  topologicalFeatures = S4Vectors::SimpleList(loops =
    S4Vectors::Pairs(GenomicRanges::GRanges(), GenomicRanges::GRanges()), borders =
    GenomicRanges::GRanges(), compartments = GenomicRanges::GRanges(), viewpoints =
    GenomicRanges::GRanges()),
  pairsFile = NULL
)
}
\arguments{
\item{file}{Path to a (m)cool file}

\item{resolution}{Resolution to use with mcool file}

\item{focus}{focus Chr. coordinates for which
interaction counts are extracted from the .(m)cool file, provided
as a character string (e.g. "II:4001-5000"). If not provided,
the entire (m)cool file will be imported.}

\item{metadata}{list of metadata}

\item{topologicalFeatures}{topologicalFeatures provided as a named SimpleList}

\item{pairsFile}{Path to an associated .pairs file}
}
\value{
a new \code{HiCExperiment} object.
}
\description{
This function has been deprecated in favor of the generic \code{HiCExperiment()}
constructor (from HiCExperiment package).
}
\examples{
library(HiContacts)
library(HiContactsData)
mcool_path <- HiContactsData::HiContactsData('yeast_wt', 'mcool')
Contacts(mcool_path, resolution = 1000)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotSaddle}
\alias{plotSaddle}
\title{Plotting saddle plots}
\usage{
plotSaddle(
  x,
  nbins = 50,
  limits = c(-1, 1),
  plotBins = FALSE,
  BPPARAM = BiocParallel::bpparam()
)
}
\arguments{
\item{x}{a HiCExperiment object with a stored \code{eigens} metadata}

\item{nbins}{Number of bins to use to discretize the eigenvectors}

\item{limits}{limits for color map being used}

\item{plotBins}{Whether to plot the distribution of bins on top of the plot}

\item{BPPARAM}{a BiocParallel registered method}
}
\value{
ggplot
}
\description{
Plotting saddle plots
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plotPs}
\alias{plotPs}
\alias{plotPsSlope}
\title{Plotting a P(s) distance law}
\usage{
plotPs(x, mapping, xlim = c(5000, 499000), ylim = c(1e-08, 1e-04))

plotPsSlope(x, mapping, xlim = c(5000, 499000), ylim = c(-3, 0))
}
\arguments{
\item{x}{the output data.frame of \code{distanceLaw} function}

\item{mapping}{aes to pass on to ggplot2}

\item{xlim}{xlim}

\item{ylim}{ylim}
}
\value{
ggplot
}
\description{
Plotting a P(s) distance law
}
\examples{
## Single P(s)

contacts_yeast <- contacts_yeast()
ps <- distanceLaw(contacts_yeast)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p))

## Comparing several P(s)

contacts_yeast <- contacts_yeast()
contacts_yeast_eco1 <- contacts_yeast_eco1()
ps_wt <- distanceLaw(contacts_yeast)
ps_wt$sample <- 'WT'
ps_eco1 <- distanceLaw(contacts_yeast_eco1)
ps_eco1$sample <- 'eco1'
ps <- rbind(ps_wt, ps_eco1)
plotPs(ps, ggplot2::aes(x = binned_distance, y = norm_p, group = sample, color = sample))
plotPsSlope(ps, ggplot2::aes(x = binned_distance, y = slope, group = sample))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{palettes}
\alias{palettes}
\alias{bwrColors}
\alias{bbrColors}
\alias{bgrColors}
\alias{afmhotrColors}
\alias{coolerColors}
\alias{rainbowColors}
\title{Matrix palettes}
\usage{
bwrColors()

bbrColors()

bgrColors()

afmhotrColors()

coolerColors()

rainbowColors()
}
\value{
A vector of colours carefully
picked for Hi-C contact heatmaps
}
\description{
Matrix palettes
}
\examples{
bwrColors()
bbrColors()
bgrColors()
afmhotrColors()
coolerColors()
rainbowColors()
}
